import logging
from sys import argv
import yaml

AL_TYPE_SPEC = {
        "float": "f",
        "int": "i",
        
        }

class CodeGen:
    def __init__(self, in_file_name, out_file_base):
        with open(in_file_name, "rb") as in_file:
            self.data = yaml.safe_load(in_file)
        self.pyx = open(out_file_base + ".pyx", "w")
        self.pyx.write("# cython: language_level=3\n\n")
        self.pxd = open(out_file_base + ".pxd", "w")
        self.pxd.write("# cython: language_level=3\n\n")

    def __del__(self):
        self.pyx.close()
        self.pxd.close()

    def generate(self):
        # Write the file prelude (if any)
        try:
            self.pyx.write(self.data["pyx_prelude"])
        except KeyError: pass
        try:
            self.pxd.write(self.data["pxd_prelude"])
        except KeyError: pass
        logging.info(f"Generating subclasses of {len(self.data['classes'])} superclasses")
        for superclass, subclasses in self.data["classes"].items():
            superclass = superclass[:-1].capitalize()
            self.gen_subclasses(superclass, subclasses)

    def gen_subclasses(self, superclass, subclasses):
        logging.info(f"Generating {len(subclasses)} subclasses of {superclass}")
        for cls_name, cls_desc in subclasses.items():
            # Write out the class declaration to the pxd file
            self.pxd.write(f"\n\ncdef class {cls_name}{superclass}({superclass}): pass")
            # Write the start of the definition to the pyx file
            self.pyx.write(f"""\n
cdef class {cls_name}{superclass}({superclass}):
    {repr(cls_desc.get('doc', ''))}

    cdef void init_with_id(self, EfxExtension efx, al.ALuint id):
        {superclass}.init_with_id(self, efx, id)
        self.type = "{cls_desc["al_effect"].upper()}"
    """)
        # Define the properties
        for prop in cls_desc["properties"]:
            self.gen_prop(superclass=superclass, subclass=cls_name, **prop)

    def gen_prop(self, name, type, superclass, subclass, doc=""):
        logging.info(f"Generating {type} property {name}")
        if type == "v3f":
            # We have to handle this differently
            self.pyx.write(f"""
    @property
    def {name}(self):
        {repr(doc)}
        cdef al.ALfloat x, y, z
        self.efx.alGet{superclass}3f(self.id, self.efx.AL_{subclass.upper()}_{name.upper()}, &x, &y, &z)
        check_al_error()
        return V3f(x, y, z)

    @{name}.setter
    def {name}(self, val):
        self.efx.al{superclass}3f(self.id, self.efx.AL_{subclass.upper()}_{name.upper()}, val[0], val[1], val[2])
        check_al_error()
            """)
        else:
            self.pyx.write(f"""
    @property
    def {name}(self):
        {repr(doc)}
        cdef al.AL{type} val
        self.efx.alGet{superclass}{AL_TYPE_SPEC[type]}(self.id, self.efx.AL_{subclass.upper()}_{name.upper()}, &val)
        check_al_error()
        return val

    @{name}.setter
    def {name}(self, val):
        self.efx.al{superclass}{AL_TYPE_SPEC[type]}(self.id, self.efx.AL_{subclass.upper()}_{name.upper()}, val)
        check_al_error()
            """)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    in_file_name = argv[1] if len(argv) > 1 else "efx_classes.yaml"
    out_file_base = argv[2] if len(argv) > 2 else "cyal/efx"
    cg = CodeGen(in_file_name, out_file_base)
    cg.generate()



# cython: language_level=3

from libc.stddef cimport size_t
from libc.string cimport strlen
from libc cimport math

from . cimport al, alc
from .exceptions cimport InvalidAlEnumError, UnsupportedExtensionError

cdef list alc_string_to_list(const alc.ALCchar* str):
    cdef:
        list specs = []
        size_t length
    while True:
        length = strlen(str)
        if length == 0:
            return specs
        specs.append(str[:length].decode("utf8"))
        str += length + 1

def get_device_specifiers():
    cdef const alc.ALCchar* specs = alc.alcGetString(NULL, alc.ALC_DEVICE_SPECIFIER)
    return alc_string_to_list(specs) if specs is not NULL else []

def get_default_device_specifier():
    cdef const alc.ALCchar* spec = alc.alcGetString(NULL, alc.ALC_DEFAULT_DEVICE_SPECIFIER)
    return spec.decode("utf8")

def get_all_device_specifiers(*, emulate=True):
    cdef alc.ALCenum enum_val = alc.alcGetEnumValue(NULL, b"ALC_ALL_DEVICES_SPECIFIER")
    cdef const alc.ALCchar *specs
    if enum_val != al.AL_NONE:
        specs = alc.alcGetString(NULL, enum_val)
        return alc_string_to_list(specs) if specs is not NULL else []
    elif emulate:
        return get_device_specifiers()
    else:
        raise UnsupportedExtensionError("ALC_ENUMERATE_ALL_EXT")

def get_default_all_device_specifier(*, emulate=True):
    cdef alc.ALCenum enum_val = alc.alcGetEnumValue(NULL, b"ALC_DEFAULT_ALL_DEVICES_SPECIFIER")
    cdef const alc.ALCchar *spec
    if enum_val != al.AL_NONE:
        spec = alc.alcGetString(NULL, enum_val)
        return spec.decode("utf8")
    elif emulate:
        return get_default_device_specifier()
    else:
        raise UnsupportedExtensionError("ALC_ENUMERATE_ALL_EXT")

def get_supported_extensions():
    cdef const alc.ALCchar* exts = alc.alcGetString(NULL, alc.ALC_EXTENSIONS)
    return exts.decode("utf8").split(' ') if exts is not NULL else []

def get_version():
    cdef alc.ALCint major, minor
    alc.alcGetIntegerv(NULL, alc.ALC_MAJOR_VERSION, 1, &major)
    alc.alcGetIntegerv(NULL, alc.ALC_MINOR_VERSION, 1, &minor)
    return (major, minor)

cdef class V3f:
    def __cinit__(self, x, y, z):
        self.data[0] = x
        self.data[1] = y
        self.data[2] = z

    def __repr__(self):
        return f"({self.data[0]}, {self.data[1]}, {self.data[2]})"

    def __getitem__(self, idx):
        if idx > 2:
            raise IndexError("vector index out of range")
        return self.data[idx]

    def __setitem__(self, idx, val):
        if idx > 2:
            raise IndexError("vector assignment index out of range")
        self.data[idx] = val

    @property
    def length(self):
        return math.hypot(math.hypot(self.data[0], self.data[1]), self.data[2])

    @property
    def x(self):
        return self.data[0]

    @x.setter
    def x(self, val):
        self.data[0] = val

    @property
    def y(self):
        return self.data[1]

    @y.setter
    def y(self, val):
        self.data[1] = val

    @property
    def z(self):
        return self.data[2]

    @z.setter
    def z(self, val):
        self.data[2] = val

cdef al.ALenum get_al_enum(str name):
    cdef bytes e_name = b"AL_" + name.upper().encode("utf8")
    cdef al.ALenum val = al.alGetEnumValue(<const al.ALchar *>e_name)
    if val == al.AL_NONE:
        raise InvalidAlEnumError()
    return val



# cython: language_level=3

from . cimport al, alc

cdef class CyalError(Exception):
    pass

# ALC_* errors

cdef class AlcError(CyalError):
    pass

cdef class DeviceNotFoundError(AlcError):
    cdef readonly bytes device_name

cdef class InvalidDeviceError(AlcError):
    pass

cdef class InvalidContextError(AlcError):
    pass

cdef class InvalidAlcEnumError(AlcError):
    pass

cdef class InvalidAlcValueError(AlcError):
    pass

cdef class UnknownAlcError(AlcError):
    cdef readonly alc.ALCenum errcode

cdef check_alc_error(alc.ALCdevice* dev)

# AL_* errors

cdef class AlError(CyalError):
    pass

cdef class InvalidNameError(AlError):
    pass

cdef class InvalidOperationError(AlError):
    pass

cdef class InvalidAlEnumError(AlError):
    pass

cdef class InvalidAlValueError(AlError):
    pass

cdef class UnknownAlError(AlError):
    cdef readonly al.ALenum errcode

cdef check_al_error()

# Cyal-specific errors

cdef class UnsupportedExtensionError(CyalError):
    cdef readonly str extension_name



# cython: language_level=3

from . cimport alc

cdef class CaptureExtension:
    cdef object __weakref__
    cdef alc.ALCenum DEFAULT_DEVICE_SPECIFIER
    cdef alc.ALCenum DEVICE_SPECIFIER
    cdef alc.ALCenum SAMPLES

cdef class CaptureDevice:
    cdef object __weakref__
    cdef readonly CaptureExtension capture_extension
    cdef alc.ALCdevice* _device
    cdef alc.ALCsizei bytes_per_sample

    @staticmethod
    cdef CaptureDevice from_device(CaptureExtension, alc.ALCdevice*, alc.ALCsizei)



# cython: language_level=3

from contextlib import contextmanager

from .buffer cimport BufferFormat
from .exceptions cimport DeviceNotFoundError, UnsupportedExtensionError, check_alc_error
from .util cimport alc_string_to_list
from . cimport al, alc

cdef class CaptureExtension:
    def __cinit__(self):
        if alc.alcIsExtensionPresent(NULL, b"ALC_EXT_CAPTURE") == al.AL_FALSE:
            raise UnsupportedExtensionError("ALC_EXT_CAPTURE")
        self.DEFAULT_DEVICE_SPECIFIER = get_enum_value("ALC_CAPTURE_DEFAULT_DEVICE_SPECIFIER")
        self.DEVICE_SPECIFIER = get_enum_value("ALC_CAPTURE_DEVICE_SPECIFIER")
        self.SAMPLES = get_enum_value("ALC_CAPTURE_SAMPLES")

    @property
    def devices(self):
        cdef const alc.ALCchar* specs = alc.alcGetString(NULL, self.DEVICE_SPECIFIER)
        return alc_string_to_list(specs) if specs is not NULL else []

    @property
    def default_device(self):
        cdef const alc.ALCchar* spec = alc.alcGetString(NULL, self.DEFAULT_DEVICE_SPECIFIER)
        return spec if spec is not NULL else None

    def open_device(self, name=None, *, format=BufferFormat.MONO16, sample_rate=48000, buf_size=512):
        cdef alc.ALCdevice* dev = alc.alcCaptureOpenDevice(<const alc.ALCchar *>name if name is not None else NULL, sample_rate, int(format), buf_size)
        if dev is NULL:
            raise DeviceNotFoundError(device_name=name)

        cdef alc.ALCsizei bps
        if format == BufferFormat.MONO8: bps = 1
        elif format == BufferFormat.MONO16: bps = 2
        elif format == BufferFormat.STEREO8: bps = 2
        elif format == BufferFormat.STEREO16: bps = 4
        else: raise RuntimeError(f"Unknown frame size for format {format}")

        return CaptureDevice.from_device(self, dev, bps)

cdef class CaptureDevice:
    def __cinit__(self):
        pass

    @staticmethod
    cdef CaptureDevice from_device(CaptureExtension ext, alc.ALCdevice* dev, alc.ALCsizei bytes_per_sample):
        cdef CaptureDevice device = CaptureDevice.__new__(CaptureDevice)
        device.capture_extension = ext
        device._device = dev
        device.bytes_per_sample = bytes_per_sample
        return device

    def __dealloc__(self):
        alc.alcCaptureCloseDevice(self._device)

    def start(self):
        alc.alcCaptureStart(self._device)
        check_alc_error(self._device)

    def stop(self):
        alc.alcCaptureStop(self._device)
        check_alc_error(self._device)

    @contextmanager
    def capturing(self):
        alc.alcCaptureStart(self._device)
        check_alc_error(self._device)
        try:
            yield self
        finally:
            alc.alcCaptureStop(self._device)
            check_alc_error(self._device)

    def capture_samples(self, buf):
        cdef al.ALubyte[:] view = buf
        if view.size == 0: return
        alc.alcCaptureSamples(self._device, &view[0], view.size / self.bytes_per_sample)
        check_alc_error(self._device)

    @property
    def name(self):
        cdef const alc.ALCchar* spec = alc.alcGetString(self._device, self.capture_extension.DEVICE_SPECIFIER)
        return spec if spec is not NULL else None

    @property
    def available_samples(self):
        cdef al.ALint val
        alc.alcGetIntegerv(self._device, self.capture_extension.SAMPLES, 1, &val)
        return val

cdef alc.ALCenum get_enum_value(const alc.ALCchar* name):
    cdef alc.ALCenum val = alc.alcGetEnumValue(NULL, name)
    if val == al.AL_NONE:
        raise RuntimeError(f"Could not get enum value for {name}")
    return val



# cython: language_level=3

from . cimport al, alc

cdef class CyalError(Exception):
    pass

cdef class AlcError(CyalError):
    pass

cdef class DeviceNotFoundError(AlcError):
    def __cinit__(self, *args, **kwargs):
        self.device_name = kwargs.get("device_name", None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args)

    def __str__(self):
        if self.device_name is None:
            return "No OpenAL devices found"
        else:
            return f"OpenAL device {self.device_name} not found"

cdef class InvalidDeviceError(AlcError):
    @property
    def errcode(self):
        return alc.ALC_INVALID_DEVICE

    def __str__(self):
        return "Invalid OpenAL device"

cdef class InvalidContextError(AlcError):
    @property
    def errcode(self):
        return alc.ALC_INVALID_CONTEXT

    def __str__(self):
        return "Invalid OpenAL context"

cdef class InvalidAlcEnumError(AlcError):
    @property
    def errcode(self):
        return alc.ALC_INVALID_ENUM

    def __str__(self):
        return "Invalid OpenAL context enum value"

cdef class InvalidAlcValueError(AlcError):
    @property
    def errcode(self):
        return alc.ALC_INVALID_VALUE

    def __str__(self):
        return "Invalid OpenAL context parameter value"

cdef class UnknownAlcError(AlcError):
    def __cinit__(self, *args, alc_errcode, **kwargs):
        self.errcode = alc_errcode

    def __init__(self, *args, alc_errcode, **kwargs):
        super().__init__(*args)

    def __str__(self):
        return f"Unknown OpenAL context error (code {self.errcode:X})"

cdef check_alc_error(alc.ALCdevice *dev):
    cdef alc.ALCenum errcode = alc.alcGetError(dev)
    if errcode == alc.ALC_NO_ERROR:
        return # Fast path
    elif errcode == alc.ALC_INVALID_DEVICE:
        raise InvalidDeviceError()
    elif errcode == alc.ALC_INVALID_CONTEXT:
        raise InvalidContextError()
    elif errcode == alc.ALC_INVALID_ENUM:
        raise InvalidAlcEnumError()
    elif errcode == alc.ALC_INVALID_VALUE:
        raise InvalidAlcValueError()
    elif errcode == alc.ALC_OUT_OF_MEMORY:
        raise MemoryError()
    else:
        raise UnknownAlcError(alc_errcode=errcode)

cdef class AlError(CyalError):
    pass

cdef class InvalidNameError(AlError):
    @property
    def errcode(self):
        return al.AL_INVALID_NAME

    def __str__(self):
        cdef const al.ALchar* string = al.alGetString(al.AL_INVALID_NAME)
        return string.decode() if string is not NULL else "Invalid OpenAL object name"

cdef class InvalidOperationError(AlError):
    @property
    def errcode(self):
        return al.AL_INVALID_OPERATION

    def __str__(self):
        cdef const al.ALchar* string = al.alGetString(al.AL_INVALID_OPERATION)
        return string.decode() if string is not NULL else "Invalid OpenAL operation"

cdef class InvalidAlEnumError(AlError):
    @property
    def errcode(self):
        return al.AL_INVALID_ENUM

    def __str__(self):
        cdef const al.ALchar* string = al.alGetString(al.AL_INVALID_ENUM)
        return string.decode() if string is not NULL else "Invalid OpenAL enum value"

cdef class InvalidAlValueError(AlError):
    @property
    def errcode(self):
        return al.AL_INVALID_VALUE

    def __str__(self):
        cdef const al.ALchar* string = al.alGetString(al.AL_INVALID_VALUE)
        return string.decode() if string is not NULL else "Invalid OpenAL parameter value"

cdef class UnknownAlError(AlError):
    def __cinit__(self, *args, al_errcode, **kwargs):
        self.errcode = al_errcode

    def __init__(self, *args, alc_errcode, **kwargs):
        super().__init__(*args)

    def __str__(self):
        cdef const al.ALchar* string = al.alGetString(self.errcode)
        return string.decode() if string is not NULL else f"Unknown OpenAL error (code {self.errcode:X})"

cdef check_al_error():
    cdef al.ALenum errcode = al.alGetError()
    if errcode == al.AL_NO_ERROR:
        return # Fast path
    elif errcode == al.AL_INVALID_NAME:
        raise InvalidNameError()
    elif errcode == al.AL_INVALID_ENUM:
        raise InvalidAlEnumError()
    elif errcode == al.AL_INVALID_VALUE:
        raise InvalidAlValueError()
    elif errcode == al.AL_INVALID_OPERATION:
        raise InvalidOperationError()
    elif errcode == al.AL_OUT_OF_MEMORY:
        raise MemoryError()
    else:
        raise UnknownAlError(al_errcode=errcode)

cdef class UnsupportedExtensionError(CyalError):
    def __cinit__(self, ext_name):
        self.extension_name = ext_name

    def __str__(self):
        return f"Unsupported OpenAL extension: {self.extension_name}"



from . cimport al

cdef extern from "efx.h":
    cdef enum:
        ALC_EXT_EFX_NAME

        ALC_EFX_MAJOR_VERSION                   
        ALC_EFX_MINOR_VERSION                   
        ALC_MAX_AUXILIARY_SENDS                 


        # Listener properties. 
        AL_METERS_PER_UNIT                      

        # Source properties. 
        AL_DIRECT_FILTER                        
        AL_AUXILIARY_SEND_FILTER                
        AL_AIR_ABSORPTION_FACTOR                
        AL_ROOM_ROLLOFF_FACTOR                  
        AL_CONE_OUTER_GAINHF                    
        AL_DIRECT_FILTER_GAINHF_AUTO            
        AL_AUXILIARY_SEND_FILTER_GAIN_AUTO      
        AL_AUXILIARY_SEND_FILTER_GAINHF_AUTO    


        # Effect properties. 

        # Reverb effect parameters 
        AL_REVERB_DENSITY                       
        AL_REVERB_DIFFUSION                     
        AL_REVERB_GAIN                          
        AL_REVERB_GAINHF                        
        AL_REVERB_DECAY_TIME                    
        AL_REVERB_DECAY_HFRATIO                 
        AL_REVERB_REFLECTIONS_GAIN              
        AL_REVERB_REFLECTIONS_DELAY             
        AL_REVERB_LATE_REVERB_GAIN              
        AL_REVERB_LATE_REVERB_DELAY             
        AL_REVERB_AIR_ABSORPTION_GAINHF         
        AL_REVERB_ROOM_ROLLOFF_FACTOR           
        AL_REVERB_DECAY_HFLIMIT                 

        # EAX Reverb effect parameters 
        AL_EAXREVERB_DENSITY                    
        AL_EAXREVERB_DIFFUSION                  
        AL_EAXREVERB_GAIN                       
        AL_EAXREVERB_GAINHF                     
        AL_EAXREVERB_GAINLF                     
        AL_EAXREVERB_DECAY_TIME                 
        AL_EAXREVERB_DECAY_HFRATIO              
        AL_EAXREVERB_DECAY_LFRATIO              
        AL_EAXREVERB_REFLECTIONS_GAIN           
        AL_EAXREVERB_REFLECTIONS_DELAY          
        AL_EAXREVERB_REFLECTIONS_PAN            
        AL_EAXREVERB_LATE_REVERB_GAIN           
        AL_EAXREVERB_LATE_REVERB_DELAY          
        AL_EAXREVERB_LATE_REVERB_PAN            
        AL_EAXREVERB_ECHO_TIME                  
        AL_EAXREVERB_ECHO_DEPTH                 
        AL_EAXREVERB_MODULATION_TIME            
        AL_EAXREVERB_MODULATION_DEPTH           
        AL_EAXREVERB_AIR_ABSORPTION_GAINHF      
        AL_EAXREVERB_HFREFERENCE                
        AL_EAXREVERB_LFREFERENCE                
        AL_EAXREVERB_ROOM_ROLLOFF_FACTOR        
        AL_EAXREVERB_DECAY_HFLIMIT              

        # Chorus effect parameters 
        AL_CHORUS_WAVEFORM                      
        AL_CHORUS_PHASE                         
        AL_CHORUS_RATE                          
        AL_CHORUS_DEPTH                         
        AL_CHORUS_FEEDBACK                      
        AL_CHORUS_DELAY                         

        # Distortion effect parameters 
        AL_DISTORTION_EDGE                      
        AL_DISTORTION_GAIN                      
        AL_DISTORTION_LOWPASS_CUTOFF            
        AL_DISTORTION_EQCENTER                  
        AL_DISTORTION_EQBANDWIDTH               

        # Echo effect parameters 
        AL_ECHO_DELAY                           
        AL_ECHO_LRDELAY                         
        AL_ECHO_DAMPING                         
        AL_ECHO_FEEDBACK                        
        AL_ECHO_SPREAD                          

        # Flanger effect parameters 
        AL_FLANGER_WAVEFORM                     
        AL_FLANGER_PHASE                        
        AL_FLANGER_RATE                         
        AL_FLANGER_DEPTH                        
        AL_FLANGER_FEEDBACK                     
        AL_FLANGER_DELAY                        

        # Frequency shifter effect parameters 
        AL_FREQUENCY_SHIFTER_FREQUENCY          
        AL_FREQUENCY_SHIFTER_LEFT_DIRECTION     
        AL_FREQUENCY_SHIFTER_RIGHT_DIRECTION    

        # Vocal morpher effect parameters 
        AL_VOCAL_MORPHER_PHONEMEA               
        AL_VOCAL_MORPHER_PHONEMEA_COARSE_TUNING 
        AL_VOCAL_MORPHER_PHONEMEB               
        AL_VOCAL_MORPHER_PHONEMEB_COARSE_TUNING 
        AL_VOCAL_MORPHER_WAVEFORM               
        AL_VOCAL_MORPHER_RATE                   

        # Pitchshifter effect parameters 
        AL_PITCH_SHIFTER_COARSE_TUNE            
        AL_PITCH_SHIFTER_FINE_TUNE              

        # Ringmodulator effect parameters 
        AL_RING_MODULATOR_FREQUENCY             
        AL_RING_MODULATOR_HIGHPASS_CUTOFF       
        AL_RING_MODULATOR_WAVEFORM              

        # Autowah effect parameters 
        AL_AUTOWAH_ATTACK_TIME                  
        AL_AUTOWAH_RELEASE_TIME                 
        AL_AUTOWAH_RESONANCE                    
        AL_AUTOWAH_PEAK_GAIN                    

        # Compressor effect parameters 
        AL_COMPRESSOR_ONOFF                     

        # Equalizer effect parameters 
        AL_EQUALIZER_LOW_GAIN                   
        AL_EQUALIZER_LOW_CUTOFF                 
        AL_EQUALIZER_MID1_GAIN                  
        AL_EQUALIZER_MID1_CENTER                
        AL_EQUALIZER_MID1_WIDTH                 
        AL_EQUALIZER_MID2_GAIN                  
        AL_EQUALIZER_MID2_CENTER                
        AL_EQUALIZER_MID2_WIDTH                 
        AL_EQUALIZER_HIGH_GAIN                  
        AL_EQUALIZER_HIGH_CUTOFF                

        # Effect type 
        AL_EFFECT_FIRST_PARAMETER               
        AL_EFFECT_LAST_PARAMETER                
        AL_EFFECT_TYPE                          

        # Effect types, used with the AL_EFFECT_TYPE property 
        AL_EFFECT_NULL                          
        AL_EFFECT_REVERB                        
        AL_EFFECT_CHORUS                        
        AL_EFFECT_DISTORTION                    
        AL_EFFECT_ECHO                          
        AL_EFFECT_FLANGER                       
        AL_EFFECT_FREQUENCY_SHIFTER             
        AL_EFFECT_VOCAL_MORPHER                 
        AL_EFFECT_PITCH_SHIFTER                 
        AL_EFFECT_RING_MODULATOR                
        AL_EFFECT_AUTOWAH                       
        AL_EFFECT_COMPRESSOR                    
        AL_EFFECT_EQUALIZER                     
        AL_EFFECT_EAXREVERB                     

        # Auxiliary Effect Slot properties. 
        AL_EFFECTSLOT_EFFECT                    
        AL_EFFECTSLOT_GAIN                      
        AL_EFFECTSLOT_AUXILIARY_SEND_AUTO       

        # NULL Auxiliary Slot ID to disable a source send. 
        AL_EFFECTSLOT_NULL                      


        # Filter properties. 

        # Lowpass filter parameters 
        AL_LOWPASS_GAIN                         
        AL_LOWPASS_GAINHF                       

        # Highpass filter parameters 
        AL_HIGHPASS_GAIN                        
        AL_HIGHPASS_GAINLF                      

        # Bandpass filter parameters 
        AL_BANDPASS_GAIN                        
        AL_BANDPASS_GAINLF                      
        AL_BANDPASS_GAINHF                      

        # Filter type 
        AL_FILTER_FIRST_PARAMETER               
        AL_FILTER_LAST_PARAMETER                
        AL_FILTER_TYPE                          

        # Filter types, used with the AL_FILTER_TYPE property 
        AL_FILTER_NULL                          
        AL_FILTER_LOWPASS                       
        AL_FILTER_HIGHPASS                      
        AL_FILTER_BANDPASS                      

        # Filter ranges and defaults. 

        # Lowpass filter 
        AL_LOWPASS_MIN_GAIN                     
        AL_LOWPASS_MAX_GAIN                     
        AL_LOWPASS_DEFAULT_GAIN                 

        AL_LOWPASS_MIN_GAINHF                   
        AL_LOWPASS_MAX_GAINHF                   
        AL_LOWPASS_DEFAULT_GAINHF               

        # Highpass filter 
        AL_HIGHPASS_MIN_GAIN                    
        AL_HIGHPASS_MAX_GAIN                    
        AL_HIGHPASS_DEFAULT_GAIN                

        AL_HIGHPASS_MIN_GAINLF                  
        AL_HIGHPASS_MAX_GAINLF                  
        AL_HIGHPASS_DEFAULT_GAINLF              

        # Bandpass filter 
        AL_BANDPASS_MIN_GAIN                    
        AL_BANDPASS_MAX_GAIN                    
        AL_BANDPASS_DEFAULT_GAIN                

        AL_BANDPASS_MIN_GAINHF                  
        AL_BANDPASS_MAX_GAINHF                  
        AL_BANDPASS_DEFAULT_GAINHF              

        AL_BANDPASS_MIN_GAINLF                  
        AL_BANDPASS_MAX_GAINLF                  
        AL_BANDPASS_DEFAULT_GAINLF              


        # Effect parameter ranges and defaults. 

        # Standard reverb effect 
        AL_REVERB_MIN_DENSITY                   
        AL_REVERB_MAX_DENSITY                   
        AL_REVERB_DEFAULT_DENSITY               

        AL_REVERB_MIN_DIFFUSION                 
        AL_REVERB_MAX_DIFFUSION                 
        AL_REVERB_DEFAULT_DIFFUSION             

        AL_REVERB_MIN_GAIN                      
        AL_REVERB_MAX_GAIN                      
        AL_REVERB_DEFAULT_GAIN                  

        AL_REVERB_MIN_GAINHF                    
        AL_REVERB_MAX_GAINHF                    
        AL_REVERB_DEFAULT_GAINHF                

        AL_REVERB_MIN_DECAY_TIME                
        AL_REVERB_MAX_DECAY_TIME                
        AL_REVERB_DEFAULT_DECAY_TIME            

        AL_REVERB_MIN_DECAY_HFRATIO             
        AL_REVERB_MAX_DECAY_HFRATIO             
        AL_REVERB_DEFAULT_DECAY_HFRATIO         

        AL_REVERB_MIN_REFLECTIONS_GAIN          
        AL_REVERB_MAX_REFLECTIONS_GAIN          
        AL_REVERB_DEFAULT_REFLECTIONS_GAIN      

        AL_REVERB_MIN_REFLECTIONS_DELAY         
        AL_REVERB_MAX_REFLECTIONS_DELAY         
        AL_REVERB_DEFAULT_REFLECTIONS_DELAY     

        AL_REVERB_MIN_LATE_REVERB_GAIN          
        AL_REVERB_MAX_LATE_REVERB_GAIN          
        AL_REVERB_DEFAULT_LATE_REVERB_GAIN      

        AL_REVERB_MIN_LATE_REVERB_DELAY         
        AL_REVERB_MAX_LATE_REVERB_DELAY         
        AL_REVERB_DEFAULT_LATE_REVERB_DELAY     

        AL_REVERB_MIN_AIR_ABSORPTION_GAINHF     
        AL_REVERB_MAX_AIR_ABSORPTION_GAINHF     
        AL_REVERB_DEFAULT_AIR_ABSORPTION_GAINHF 

        AL_REVERB_MIN_ROOM_ROLLOFF_FACTOR       
        AL_REVERB_MAX_ROOM_ROLLOFF_FACTOR       
        AL_REVERB_DEFAULT_ROOM_ROLLOFF_FACTOR   

        AL_REVERB_MIN_DECAY_HFLIMIT             
        AL_REVERB_MAX_DECAY_HFLIMIT             
        AL_REVERB_DEFAULT_DECAY_HFLIMIT         

        # EAX reverb effect 
        AL_EAXREVERB_MIN_DENSITY                
        AL_EAXREVERB_MAX_DENSITY                
        AL_EAXREVERB_DEFAULT_DENSITY            

        AL_EAXREVERB_MIN_DIFFUSION              
        AL_EAXREVERB_MAX_DIFFUSION              
        AL_EAXREVERB_DEFAULT_DIFFUSION          

        AL_EAXREVERB_MIN_GAIN                   
        AL_EAXREVERB_MAX_GAIN                   
        AL_EAXREVERB_DEFAULT_GAIN               

        AL_EAXREVERB_MIN_GAINHF                 
        AL_EAXREVERB_MAX_GAINHF                 
        AL_EAXREVERB_DEFAULT_GAINHF             

        AL_EAXREVERB_MIN_GAINLF                 
        AL_EAXREVERB_MAX_GAINLF                 
        AL_EAXREVERB_DEFAULT_GAINLF             

        AL_EAXREVERB_MIN_DECAY_TIME             
        AL_EAXREVERB_MAX_DECAY_TIME             
        AL_EAXREVERB_DEFAULT_DECAY_TIME         

        AL_EAXREVERB_MIN_DECAY_HFRATIO          
        AL_EAXREVERB_MAX_DECAY_HFRATIO          
        AL_EAXREVERB_DEFAULT_DECAY_HFRATIO      

        AL_EAXREVERB_MIN_DECAY_LFRATIO          
        AL_EAXREVERB_MAX_DECAY_LFRATIO          
        AL_EAXREVERB_DEFAULT_DECAY_LFRATIO      

        AL_EAXREVERB_MIN_REFLECTIONS_GAIN       
        AL_EAXREVERB_MAX_REFLECTIONS_GAIN       
        AL_EAXREVERB_DEFAULT_REFLECTIONS_GAIN   

        AL_EAXREVERB_MIN_REFLECTIONS_DELAY      
        AL_EAXREVERB_MAX_REFLECTIONS_DELAY      
        AL_EAXREVERB_DEFAULT_REFLECTIONS_DELAY  

        AL_EAXREVERB_DEFAULT_REFLECTIONS_PAN_XYZ

        AL_EAXREVERB_MIN_LATE_REVERB_GAIN       
        AL_EAXREVERB_MAX_LATE_REVERB_GAIN       
        AL_EAXREVERB_DEFAULT_LATE_REVERB_GAIN   

        AL_EAXREVERB_MIN_LATE_REVERB_DELAY      
        AL_EAXREVERB_MAX_LATE_REVERB_DELAY      
        AL_EAXREVERB_DEFAULT_LATE_REVERB_DELAY  

        AL_EAXREVERB_DEFAULT_LATE_REVERB_PAN_XYZ

        AL_EAXREVERB_MIN_ECHO_TIME              
        AL_EAXREVERB_MAX_ECHO_TIME              
        AL_EAXREVERB_DEFAULT_ECHO_TIME          

        AL_EAXREVERB_MIN_ECHO_DEPTH             
        AL_EAXREVERB_MAX_ECHO_DEPTH             
        AL_EAXREVERB_DEFAULT_ECHO_DEPTH         

        AL_EAXREVERB_MIN_MODULATION_TIME        
        AL_EAXREVERB_MAX_MODULATION_TIME        
        AL_EAXREVERB_DEFAULT_MODULATION_TIME    

        AL_EAXREVERB_MIN_MODULATION_DEPTH       
        AL_EAXREVERB_MAX_MODULATION_DEPTH       
        AL_EAXREVERB_DEFAULT_MODULATION_DEPTH   

        AL_EAXREVERB_MIN_AIR_ABSORPTION_GAINHF  
        AL_EAXREVERB_MAX_AIR_ABSORPTION_GAINHF  
        AL_EAXREVERB_DEFAULT_AIR_ABSORPTION_GAINHF

        AL_EAXREVERB_MIN_HFREFERENCE            
        AL_EAXREVERB_MAX_HFREFERENCE            
        AL_EAXREVERB_DEFAULT_HFREFERENCE        

        AL_EAXREVERB_MIN_LFREFERENCE            
        AL_EAXREVERB_MAX_LFREFERENCE            
        AL_EAXREVERB_DEFAULT_LFREFERENCE        

        AL_EAXREVERB_MIN_ROOM_ROLLOFF_FACTOR    
        AL_EAXREVERB_MAX_ROOM_ROLLOFF_FACTOR    
        AL_EAXREVERB_DEFAULT_ROOM_ROLLOFF_FACTOR

        AL_EAXREVERB_MIN_DECAY_HFLIMIT          
        AL_EAXREVERB_MAX_DECAY_HFLIMIT          
        AL_EAXREVERB_DEFAULT_DECAY_HFLIMIT      

        # Chorus effect 
        AL_CHORUS_WAVEFORM_SINUSOID             
        AL_CHORUS_WAVEFORM_TRIANGLE             

        AL_CHORUS_MIN_WAVEFORM                  
        AL_CHORUS_MAX_WAVEFORM                  
        AL_CHORUS_DEFAULT_WAVEFORM              

        AL_CHORUS_MIN_PHASE                     
        AL_CHORUS_MAX_PHASE                     
        AL_CHORUS_DEFAULT_PHASE                 

        AL_CHORUS_MIN_RATE                      
        AL_CHORUS_MAX_RATE                      
        AL_CHORUS_DEFAULT_RATE                  

        AL_CHORUS_MIN_DEPTH                     
        AL_CHORUS_MAX_DEPTH                     
        AL_CHORUS_DEFAULT_DEPTH                 

        AL_CHORUS_MIN_FEEDBACK                  
        AL_CHORUS_MAX_FEEDBACK                  
        AL_CHORUS_DEFAULT_FEEDBACK              

        AL_CHORUS_MIN_DELAY                     
        AL_CHORUS_MAX_DELAY                     
        AL_CHORUS_DEFAULT_DELAY                 

        # Distortion effect 
        AL_DISTORTION_MIN_EDGE                  
        AL_DISTORTION_MAX_EDGE                  
        AL_DISTORTION_DEFAULT_EDGE              

        AL_DISTORTION_MIN_GAIN                  
        AL_DISTORTION_MAX_GAIN                  
        AL_DISTORTION_DEFAULT_GAIN              

        AL_DISTORTION_MIN_LOWPASS_CUTOFF        
        AL_DISTORTION_MAX_LOWPASS_CUTOFF        
        AL_DISTORTION_DEFAULT_LOWPASS_CUTOFF    

        AL_DISTORTION_MIN_EQCENTER              
        AL_DISTORTION_MAX_EQCENTER              
        AL_DISTORTION_DEFAULT_EQCENTER          

        AL_DISTORTION_MIN_EQBANDWIDTH           
        AL_DISTORTION_MAX_EQBANDWIDTH           
        AL_DISTORTION_DEFAULT_EQBANDWIDTH       

        # Echo effect 
        AL_ECHO_MIN_DELAY                       
        AL_ECHO_MAX_DELAY                       
        AL_ECHO_DEFAULT_DELAY                   

        AL_ECHO_MIN_LRDELAY                     
        AL_ECHO_MAX_LRDELAY                     
        AL_ECHO_DEFAULT_LRDELAY                 

        AL_ECHO_MIN_DAMPING                     
        AL_ECHO_MAX_DAMPING                     
        AL_ECHO_DEFAULT_DAMPING                 

        AL_ECHO_MIN_FEEDBACK                    
        AL_ECHO_MAX_FEEDBACK                    
        AL_ECHO_DEFAULT_FEEDBACK                

        AL_ECHO_MIN_SPREAD                      
        AL_ECHO_MAX_SPREAD                      
        AL_ECHO_DEFAULT_SPREAD                  

        # Flanger effect 
        AL_FLANGER_WAVEFORM_SINUSOID            
        AL_FLANGER_WAVEFORM_TRIANGLE            

        AL_FLANGER_MIN_WAVEFORM                 
        AL_FLANGER_MAX_WAVEFORM                 
        AL_FLANGER_DEFAULT_WAVEFORM             

        AL_FLANGER_MIN_PHASE                    
        AL_FLANGER_MAX_PHASE                    
        AL_FLANGER_DEFAULT_PHASE                

        AL_FLANGER_MIN_RATE                     
        AL_FLANGER_MAX_RATE                     
        AL_FLANGER_DEFAULT_RATE                 

        AL_FLANGER_MIN_DEPTH                    
        AL_FLANGER_MAX_DEPTH                    
        AL_FLANGER_DEFAULT_DEPTH                

        AL_FLANGER_MIN_FEEDBACK                 
        AL_FLANGER_MAX_FEEDBACK                 
        AL_FLANGER_DEFAULT_FEEDBACK             

        AL_FLANGER_MIN_DELAY                    
        AL_FLANGER_MAX_DELAY                    
        AL_FLANGER_DEFAULT_DELAY                

        # Frequency shifter effect 
        AL_FREQUENCY_SHIFTER_MIN_FREQUENCY      
        AL_FREQUENCY_SHIFTER_MAX_FREQUENCY      
        AL_FREQUENCY_SHIFTER_DEFAULT_FREQUENCY  

        AL_FREQUENCY_SHIFTER_MIN_LEFT_DIRECTION 
        AL_FREQUENCY_SHIFTER_MAX_LEFT_DIRECTION 
        AL_FREQUENCY_SHIFTER_DEFAULT_LEFT_DIRECTION

        AL_FREQUENCY_SHIFTER_DIRECTION_DOWN     
        AL_FREQUENCY_SHIFTER_DIRECTION_UP       
        AL_FREQUENCY_SHIFTER_DIRECTION_OFF      

        AL_FREQUENCY_SHIFTER_MIN_RIGHT_DIRECTION
        AL_FREQUENCY_SHIFTER_MAX_RIGHT_DIRECTION
        AL_FREQUENCY_SHIFTER_DEFAULT_RIGHT_DIRECTION

        # Vocal morpher effect 
        AL_VOCAL_MORPHER_MIN_PHONEMEA           
        AL_VOCAL_MORPHER_MAX_PHONEMEA           
        AL_VOCAL_MORPHER_DEFAULT_PHONEMEA       

        AL_VOCAL_MORPHER_MIN_PHONEMEA_COARSE_TUNING
        AL_VOCAL_MORPHER_MAX_PHONEMEA_COARSE_TUNING
        AL_VOCAL_MORPHER_DEFAULT_PHONEMEA_COARSE_TUNING

        AL_VOCAL_MORPHER_MIN_PHONEMEB           
        AL_VOCAL_MORPHER_MAX_PHONEMEB           
        AL_VOCAL_MORPHER_DEFAULT_PHONEMEB       

        AL_VOCAL_MORPHER_MIN_PHONEMEB_COARSE_TUNING
        AL_VOCAL_MORPHER_MAX_PHONEMEB_COARSE_TUNING
        AL_VOCAL_MORPHER_DEFAULT_PHONEMEB_COARSE_TUNING

        AL_VOCAL_MORPHER_PHONEME_A              
        AL_VOCAL_MORPHER_PHONEME_E              
        AL_VOCAL_MORPHER_PHONEME_I              
        AL_VOCAL_MORPHER_PHONEME_O              
        AL_VOCAL_MORPHER_PHONEME_U              
        AL_VOCAL_MORPHER_PHONEME_AA             
        AL_VOCAL_MORPHER_PHONEME_AE             
        AL_VOCAL_MORPHER_PHONEME_AH             
        AL_VOCAL_MORPHER_PHONEME_AO             
        AL_VOCAL_MORPHER_PHONEME_EH             
        AL_VOCAL_MORPHER_PHONEME_ER             
        AL_VOCAL_MORPHER_PHONEME_IH             
        AL_VOCAL_MORPHER_PHONEME_IY             
        AL_VOCAL_MORPHER_PHONEME_UH             
        AL_VOCAL_MORPHER_PHONEME_UW             
        AL_VOCAL_MORPHER_PHONEME_B              
        AL_VOCAL_MORPHER_PHONEME_D              
        AL_VOCAL_MORPHER_PHONEME_F              
        AL_VOCAL_MORPHER_PHONEME_G              
        AL_VOCAL_MORPHER_PHONEME_J              
        AL_VOCAL_MORPHER_PHONEME_K              
        AL_VOCAL_MORPHER_PHONEME_L              
        AL_VOCAL_MORPHER_PHONEME_M              
        AL_VOCAL_MORPHER_PHONEME_N              
        AL_VOCAL_MORPHER_PHONEME_P              
        AL_VOCAL_MORPHER_PHONEME_R              
        AL_VOCAL_MORPHER_PHONEME_S              
        AL_VOCAL_MORPHER_PHONEME_T              
        AL_VOCAL_MORPHER_PHONEME_V              
        AL_VOCAL_MORPHER_PHONEME_Z              

        AL_VOCAL_MORPHER_WAVEFORM_SINUSOID      
        AL_VOCAL_MORPHER_WAVEFORM_TRIANGLE      
        AL_VOCAL_MORPHER_WAVEFORM_SAWTOOTH      

        AL_VOCAL_MORPHER_MIN_WAVEFORM           
        AL_VOCAL_MORPHER_MAX_WAVEFORM           
        AL_VOCAL_MORPHER_DEFAULT_WAVEFORM       

        AL_VOCAL_MORPHER_MIN_RATE               
        AL_VOCAL_MORPHER_MAX_RATE               
        AL_VOCAL_MORPHER_DEFAULT_RATE           

        # Pitch shifter effect 
        AL_PITCH_SHIFTER_MIN_COARSE_TUNE        
        AL_PITCH_SHIFTER_MAX_COARSE_TUNE        
        AL_PITCH_SHIFTER_DEFAULT_COARSE_TUNE    

        AL_PITCH_SHIFTER_MIN_FINE_TUNE          
        AL_PITCH_SHIFTER_MAX_FINE_TUNE          
        AL_PITCH_SHIFTER_DEFAULT_FINE_TUNE      

        # Ring modulator effect 
        AL_RING_MODULATOR_MIN_FREQUENCY         
        AL_RING_MODULATOR_MAX_FREQUENCY         
        AL_RING_MODULATOR_DEFAULT_FREQUENCY     

        AL_RING_MODULATOR_MIN_HIGHPASS_CUTOFF   
        AL_RING_MODULATOR_MAX_HIGHPASS_CUTOFF   
        AL_RING_MODULATOR_DEFAULT_HIGHPASS_CUTOFF

        AL_RING_MODULATOR_SINUSOID              
        AL_RING_MODULATOR_SAWTOOTH              
        AL_RING_MODULATOR_SQUARE                

        AL_RING_MODULATOR_MIN_WAVEFORM          
        AL_RING_MODULATOR_MAX_WAVEFORM          
        AL_RING_MODULATOR_DEFAULT_WAVEFORM      

        # Autowah effect 
        AL_AUTOWAH_MIN_ATTACK_TIME              
        AL_AUTOWAH_MAX_ATTACK_TIME              
        AL_AUTOWAH_DEFAULT_ATTACK_TIME          

        AL_AUTOWAH_MIN_RELEASE_TIME             
        AL_AUTOWAH_MAX_RELEASE_TIME             
        AL_AUTOWAH_DEFAULT_RELEASE_TIME         

        AL_AUTOWAH_MIN_RESONANCE                
        AL_AUTOWAH_MAX_RESONANCE                
        AL_AUTOWAH_DEFAULT_RESONANCE            

        AL_AUTOWAH_MIN_PEAK_GAIN                
        AL_AUTOWAH_MAX_PEAK_GAIN                
        AL_AUTOWAH_DEFAULT_PEAK_GAIN            

        # Compressor effect 
        AL_COMPRESSOR_MIN_ONOFF                 
        AL_COMPRESSOR_MAX_ONOFF                 
        AL_COMPRESSOR_DEFAULT_ONOFF             

        # Equalizer effect 
        AL_EQUALIZER_MIN_LOW_GAIN               
        AL_EQUALIZER_MAX_LOW_GAIN               
        AL_EQUALIZER_DEFAULT_LOW_GAIN           

        AL_EQUALIZER_MIN_LOW_CUTOFF             
        AL_EQUALIZER_MAX_LOW_CUTOFF             
        AL_EQUALIZER_DEFAULT_LOW_CUTOFF         

        AL_EQUALIZER_MIN_MID1_GAIN              
        AL_EQUALIZER_MAX_MID1_GAIN              
        AL_EQUALIZER_DEFAULT_MID1_GAIN          

        AL_EQUALIZER_MIN_MID1_CENTER            
        AL_EQUALIZER_MAX_MID1_CENTER            
        AL_EQUALIZER_DEFAULT_MID1_CENTER        

        AL_EQUALIZER_MIN_MID1_WIDTH             
        AL_EQUALIZER_MAX_MID1_WIDTH             
        AL_EQUALIZER_DEFAULT_MID1_WIDTH         

        AL_EQUALIZER_MIN_MID2_GAIN              
        AL_EQUALIZER_MAX_MID2_GAIN              
        AL_EQUALIZER_DEFAULT_MID2_GAIN          

        AL_EQUALIZER_MIN_MID2_CENTER            
        AL_EQUALIZER_MAX_MID2_CENTER            
        AL_EQUALIZER_DEFAULT_MID2_CENTER        

        AL_EQUALIZER_MIN_MID2_WIDTH             
        AL_EQUALIZER_MAX_MID2_WIDTH             
        AL_EQUALIZER_DEFAULT_MID2_WIDTH         

        AL_EQUALIZER_MIN_HIGH_GAIN              
        AL_EQUALIZER_MAX_HIGH_GAIN              
        AL_EQUALIZER_DEFAULT_HIGH_GAIN          

        AL_EQUALIZER_MIN_HIGH_CUTOFF            
        AL_EQUALIZER_MAX_HIGH_CUTOFF            
        AL_EQUALIZER_DEFAULT_HIGH_CUTOFF        


        # Source parameter value ranges and defaults. 
        AL_MIN_AIR_ABSORPTION_FACTOR            
        AL_MAX_AIR_ABSORPTION_FACTOR            
        AL_DEFAULT_AIR_ABSORPTION_FACTOR        

        AL_MIN_ROOM_ROLLOFF_FACTOR              
        AL_MAX_ROOM_ROLLOFF_FACTOR              
        AL_DEFAULT_ROOM_ROLLOFF_FACTOR          

        AL_MIN_CONE_OUTER_GAINHF                
        AL_MAX_CONE_OUTER_GAINHF                
        AL_DEFAULT_CONE_OUTER_GAINHF            

        AL_MIN_DIRECT_FILTER_GAINHF_AUTO        
        AL_MAX_DIRECT_FILTER_GAINHF_AUTO        
        AL_DEFAULT_DIRECT_FILTER_GAINHF_AUTO    

        AL_MIN_AUXILIARY_SEND_FILTER_GAIN_AUTO  
        AL_MAX_AUXILIARY_SEND_FILTER_GAIN_AUTO  
        AL_DEFAULT_AUXILIARY_SEND_FILTER_GAIN_AUTO

        AL_MIN_AUXILIARY_SEND_FILTER_GAINHF_AUTO
        AL_MAX_AUXILIARY_SEND_FILTER_GAINHF_AUTO
        AL_DEFAULT_AUXILIARY_SEND_FILTER_GAINHF_AUTO


        # Listener parameter value ranges and defaults. 
        AL_MIN_METERS_PER_UNIT                  
        AL_MAX_METERS_PER_UNIT                  
        AL_DEFAULT_METERS_PER_UNIT              

    void alGenEffects(al.ALsizei n, al.ALuint *effects)
    void alDeleteEffects(al.ALsizei n, const al.ALuint *effects)
    al.ALboolean alIsEffect(al.ALuint effect)
    void alEffecti(al.ALuint effect, al.ALenum param, al.ALint iValue)
    void alEffectiv(al.ALuint effect, al.ALenum param, const al.ALint *piValues)
    void alEffectf(al.ALuint effect, al.ALenum param, al.ALfloat flValue)
    void alEffectfv(al.ALuint effect, al.ALenum param, const al.ALfloat *pflValues)
    void alGetEffecti(al.ALuint effect, al.ALenum param, al.ALint *piValue)
    void alGetEffectiv(al.ALuint effect, al.ALenum param, al.ALint *piValues)
    void alGetEffectf(al.ALuint effect, al.ALenum param, al.ALfloat *pflValue)
    void alGetEffectfv(al.ALuint effect, al.ALenum param, al.ALfloat *pflValues)

    void alGenFilters(al.ALsizei n, al.ALuint *filters)
    void alDeleteFilters(al.ALsizei n, const al.ALuint *filters)
    al.ALboolean alIsFilter(al.ALuint filter)
    void alFilteri(al.ALuint filter, al.ALenum param, al.ALint iValue)
    void alFilteriv(al.ALuint filter, al.ALenum param, const al.ALint *piValues)
    void alFilterf(al.ALuint filter, al.ALenum param, al.ALfloat flValue)
    void alFilterfv(al.ALuint filter, al.ALenum param, const al.ALfloat *pflValues)
    void alGetFilteri(al.ALuint filter, al.ALenum param, al.ALint *piValue)
    void alGetFilteriv(al.ALuint filter, al.ALenum param, al.ALint *piValues)
    void alGetFilterf(al.ALuint filter, al.ALenum param, al.ALfloat *pflValue)
    void alGetFilterfv(al.ALuint filter, al.ALenum param, al.ALfloat *pflValues)

    void alGenAuxiliaryEffectSlots(al.ALsizei n, al.ALuint *effectslots)
    void alDeleteAuxiliaryEffectSlots(al.ALsizei n, const al.ALuint *effectslots)
    al.ALboolean alIsAuxiliaryEffectSlot(al.ALuint effectslot)
    void alAuxiliaryEffectSloti(al.ALuint effectslot, al.ALenum param, al.ALint iValue)
    void alAuxiliaryEffectSlotiv(al.ALuint effectslot, al.ALenum param, const al.ALint *piValues)
    void alAuxiliaryEffectSlotf(al.ALuint effectslot, al.ALenum param, al.ALfloat flValue)
    void alAuxiliaryEffectSlotfv(al.ALuint effectslot, al.ALenum param, const al.ALfloat *pflValues)
    void alGetAuxiliaryEffectSloti(al.ALuint effectslot, al.ALenum param, al.ALint *piValue)
    void alGetAuxiliaryEffectSlotiv(al.ALuint effectslot, al.ALenum param, al.ALint *piValues)
    void alGetAuxiliaryEffectSlotf(al.ALuint effectslot, al.ALenum param, al.ALfloat *pflValue)
    void alGetAuxiliaryEffectSlotfv(al.ALuint effectslot, al.ALenum param, al.ALfloat *pflValues)



cdef extern from "alc.h":
    # Opaque device handle
    ctypedef struct ALCdevice:
        pass
    # Opaque context handle
    ctypedef struct ALCcontext:
        pass

    # 8-bit boolean
    ctypedef char ALCboolean

    # character
    ctypedef char ALCchar

    # signed 8-bit integer
    ctypedef signed char ALCbyte

    # unsigned 8-bit integer
    ctypedef unsigned char ALCubyte

    # signed 16-bit integer
    ctypedef short ALCshort

    # unsigned 16-bit integer
    ctypedef unsigned short ALCushort

    # signed 32-bit integer
    ctypedef int ALCint

    # unsigned 32-bit integer
    ctypedef unsigned int ALCuint

    # non-negative 32-bit integer size
    ctypedef int ALCsizei

    # 32-bit enumeration value
    ctypedef int ALCenum

    # 32-bit IEEE-754 floating-point
    ctypedef float ALCfloat

    # 64-bit IEEE-754 floating-point
    ctypedef double ALCdouble

    # void type (for opaque pointers only)
    ctypedef void ALCvoid

    cdef enum:
        # Context attribute: <int> Hz.
        ALC_FREQUENCY = 0x1007

        # Context attribute: <int> Hz.
        ALC_REFRESH = 0x1008

        # Context attribute: AL_TRUE or AL_FALSE synchronous context?
        ALC_SYNC = 0x1009

        # Context attribute: <int> requested Mono (3D) Sources.
        ALC_MONO_SOURCES = 0x1010

        # Context attribute: <int> requested Stereo Sources.
        ALC_STEREO_SOURCES = 0x1011

        # No error.
        ALC_NO_ERROR = 0

        # Invalid device handle.
        ALC_INVALID_DEVICE = 0xA001

        # Invalid context handle.
        ALC_INVALID_CONTEXT = 0xA002

        # Invalid enumeration passed to an ALC call.
        ALC_INVALID_ENUM = 0xA003

        # Invalid value passed to an ALC call.
        ALC_INVALID_VALUE = 0xA004

        # Out of memory.
        ALC_OUT_OF_MEMORY = 0xA005


        # Runtime ALC major version.
        ALC_MAJOR_VERSION = 0x1000
        # Runtime ALC minor version.
        ALC_MINOR_VERSION = 0x1001

        # Context attribute list size.
        ALC_ATTRIBUTES_SIZE = 0x1002
        # Context attribute list properties.
        ALC_ALL_ATTRIBUTES = 0x1003

        # String for the default device specifier.
        ALC_DEFAULT_DEVICE_SPECIFIER = 0x1004
        # Device specifier string.
        #
        # If device handle is NULL, it is instead a null-character separated list of
        # strings of known device specifiers (list ends with an empty string).
        ALC_DEVICE_SPECIFIER = 0x1005
        # String for space-separated list of ALC extensions.
        ALC_EXTENSIONS = 0x1006


        # Capture extension
        ALC_EXT_CAPTURE = 1
        # Capture device specifier string.
        #
        # If device handle is NULL, it is instead a null-character separated list of
        # strings of known capture device specifiers (list ends with an empty string).
        ALC_CAPTURE_DEVICE_SPECIFIER = 0x310
        # String for the default capture device specifier.
        ALC_CAPTURE_DEFAULT_DEVICE_SPECIFIER = 0x311
        # Number of sample frames available for capture.
        ALC_CAPTURE_SAMPLES = 0x312


        # Enumerate All extension
        ALC_ENUMERATE_ALL_EXT = 1
        # String for the default extended device specifier.
        ALC_DEFAULT_ALL_DEVICES_SPECIFIER = 0x1012
        # Device's extended specifier string.
        #
        # If device handle is NULL, it is instead a null-character separated list of
        # strings of known extended device specifiers (list ends with an empty string).
        ALC_ALL_DEVICES_SPECIFIER = 0x1013


    # Context management.

    # Create and attach a context to the given device.
    ALCcontext* alcCreateContext(ALCdevice *device, const ALCint *attrlist)
    # Makes the given context the active process-wide context. Passing NULL clears
    # the active context.
    ALCboolean  alcMakeContextCurrent(ALCcontext *context)
    # Resumes processing updates for the given context.
    void        alcProcessContext(ALCcontext *context)
    # Suspends updates for the given context.
    void        alcSuspendContext(ALCcontext *context)
    # Remove a context from its device and destroys it.
    void        alcDestroyContext(ALCcontext *context)
    # Returns the currently active context.
    ALCcontext* alcGetCurrentContext()
    # Returns the device that a particular context is attached to.
    ALCdevice*  alcGetContextsDevice(ALCcontext *context)

    # Device management.

    # Opens the named playback device.
    ALCdevice* alcOpenDevice(const ALCchar *devicename)
    # Closes the given playback device.
    ALCboolean alcCloseDevice(ALCdevice *device)

    # Error support.

    # Obtain the most recent Device error.
    ALCenum alcGetError(ALCdevice *device)

    # Extension support.

    # Query for the presence of an extension on the device. Pass a NULL device to
    # query a device-inspecific extension.
    ALCboolean alcIsExtensionPresent(ALCdevice *device, const ALCchar *extname)
    # Retrieve the address of a function. Given a non-NULL device, the returned
    # function may be device-specific.
    ALCvoid*   alcGetProcAddress(ALCdevice *device, const ALCchar *funcname)
    # Retrieve the value of an enum. Given a non-NULL device, the returned value
    # may be device-specific.
    ALCenum    alcGetEnumValue(ALCdevice *device, const ALCchar *enumname)

    # Query functions.

    # Returns information about the device, and error strings.
    const ALCchar* alcGetString(ALCdevice *device, ALCenum param)
    # Returns information about the device and the version of OpenAL.
    void           alcGetIntegerv(ALCdevice *device, ALCenum param, ALCsizei size, ALCint *values)

    # Capture functions.

    # Opens the named capture device with the given frequency, format, and buffer
    # size.
    ALCdevice* alcCaptureOpenDevice(const ALCchar *devicename, ALCuint frequency, ALCenum format, ALCsizei buffersize)
    # Closes the given capture device.
    ALCboolean alcCaptureCloseDevice(ALCdevice *device)
    # Starts capturing samples into the device buffer.
    void       alcCaptureStart(ALCdevice *device)
    # Stops capturing samples. Samples in the device buffer remain available.
    void       alcCaptureStop(ALCdevice *device)
    # Reads samples from the device buffer.
    void       alcCaptureSamples(ALCdevice *device, ALCvoid *buffer, ALCsizei samples)



# cython: language_level=3

from .context cimport Context
from . cimport al

cdef class Buffer:
    cdef object __weakref__
    cdef readonly Context context
    cdef readonly al.ALuint id

    @staticmethod
    cdef Buffer from_id(Context ctx, al.ALuint id)

cpdef enum BufferFormat:
        MONO8 = al.AL_FORMAT_MONO8
        MONO16 = al.AL_FORMAT_MONO16
        STEREO8 = al.AL_FORMAT_STEREO8
        STEREO16 = al.AL_FORMAT_STEREO16



cdef extern from "al.h":
    #  8-bit boolean
    ctypedef char ALboolean

    #  character
    ctypedef char ALchar

    #  signed 8-bit integer
    ctypedef signed char ALbyte

    #  unsigned 8-bit integer
    ctypedef unsigned char ALubyte

    #  signed 16-bit integer
    ctypedef short ALshort

    #  unsigned 16-bit integer
    ctypedef unsigned short ALushort

    #  signed 32-bit integer
    ctypedef int ALint

    #  unsigned 32-bit integer
    ctypedef unsigned int ALuint

    #  non-negative 32-bit integer size
    ctypedef int ALsizei

    #  32-bit enumeration value
    ctypedef int ALenum

    #  32-bit IEEE-754 floating-point
    ctypedef float ALfloat

    #  64-bit IEEE-754 floating-point
    ctypedef double ALdouble

    #  void type (opaque pointers only)
    ctypedef void ALvoid

    cdef enum:
        #  No distance model or no buffer
        AL_NONE = 0

        #  Boolean False.
        AL_FALSE = 0

        #  Boolean True.
        AL_TRUE = 1


        # Relative source.
        # Type:    ALboolean
        # Range:   [AL_FALSE, AL_TRUE]
        # Default: AL_FALSE
        #
        # Specifies if the source uses relative coordinates.
        AL_SOURCE_RELATIVE = 0x202


        # Inner cone angle, in degrees.
        # Type:    ALint, ALfloat
        # Range:   [0 - 360]
        # Default: 360
        #
        # The angle covered by the inner cone, the area within which the source will
        # not be attenuated by direction.
        AL_CONE_INNER_ANGLE = 0x1001

        # Outer cone angle, in degrees.
        # Range:   [0 - 360]
        # Default: 360
        #
        # The angle covered by the outer cone, the area outside of which the source
        # will be fully attenuated by direction.
        AL_CONE_OUTER_ANGLE = 0x1002

        # Source pitch.
        # Type:    ALfloat
        # Range:   [0.5 - 2.0]
        # Default: 1.0
        #
        # A multiplier for the sample rate of the source's buffer.
        AL_PITCH = 0x1003

        # Source or listener position.
        # Type:    ALfloat[3], ALint[3]
        # Default: {0, 0, 0}
        #
        # The source or listener location in three dimensional space.
        #
        # OpenAL uses a right handed coordinate system, like OpenGL, where with a
        # default view, X points right (thumb), Y points up (index finger), and Z
        # points towards the viewer/camera (middle finger).
        #
        # To change from or to a left handed coordinate system, negate the Z
        # component.
        AL_POSITION = 0x1004

        # Source direction.
        # Type:    ALfloat[3], ALint[3]
        # Default: {0, 0, 0}
        #
        # Specifies the current direction in local space. A zero-length vector
        # specifies an omni-directional source (cone is ignored).
        #
        # To change from or to a left handed coordinate system, negate the Z
        # component.
        AL_DIRECTION = 0x1005

        # Source or listener velocity.
        # Type:    ALfloat[3], ALint[3]
        # Default: {0, 0, 0}
        #
        # Specifies the current velocity, relative to the position.
        #
        # To change from or to a left handed coordinate system, negate the Z
        # component.
        AL_VELOCITY = 0x1006

        # Source looping.
        # Type:    ALboolean
        # Range:   [AL_FALSE, AL_TRUE]
        # Default: AL_FALSE
        #
        # Specifies whether source playback loops.
        AL_LOOPING = 0x1007

        # Source buffer.
        # Type:    ALuint
        # Range:   any valid Buffer ID
        # Default: AL_NONE
        #
        # Specifies the buffer to provide sound samples for a source.
        AL_BUFFER = 0x1009

        # Source or listener gain.
        # Type:  ALfloat
        # Range: [0.0 - ]
        #
        # For sources, an initial linear gain value (before attenuation is applied).
        # For the listener, an output linear gain adjustment.
        #
        # A value of 1.0 means unattenuated. Each division by 2 equals an attenuation
        # of about -6dB. Each multiplication by 2 equals an amplification of about
        # +6dB.
        AL_GAIN = 0x100A

        # Minimum source gain.
        # Type:  ALfloat
        # Range: [0.0 - 1.0]
        #
        # The minimum gain allowed for a source, after distance and cone attenuation
        # are applied (if applicable).
        AL_MIN_GAIN = 0x100D

        # Maximum source gain.
        # Type:  ALfloat
        # Range: [0.0 - 1.0]
        #
        # The maximum gain allowed for a source, after distance and cone attenuation
        # are applied (if applicable).
        AL_MAX_GAIN = 0x100E

        # Listener orientation.
        # Type:    ALfloat[6]
        # Default: {0.0, 0.0, -1.0, 0.0, 1.0, 0.0}
        #
        # Effectively two three dimensional vectors. The first vector is the front (or
        # "at") and the second is the top (or "up"). Both vectors are relative to the
        # listener position.
        #
        # To change from or to a left handed coordinate system, negate the Z
        # component of both vectors.
        AL_ORIENTATION = 0x100F

        # Source state (query only).
        # Type:  ALenum
        # Range: [AL_INITIAL, AL_PLAYING, AL_PAUSED, AL_STOPPED]
        AL_SOURCE_STATE = 0x1010

        #  Source state values.
        AL_INITIAL = 0x1011
        AL_PLAYING = 0x1012
        AL_PAUSED = 0x1013
        AL_STOPPED = 0x1014

        # Source Buffer Queue size (query only).
        # Type: ALint
        #
        # The number of buffers queued using alSourceQueueBuffers, minus the buffers
        # removed with alSourceUnqueueBuffers.
        AL_BUFFERS_QUEUED = 0x1015

        # Source Buffer Queue processed count (query only).
        # Type: ALint
        #
        # The number of queued buffers that have been fully processed, and can be
        # removed with alSourceUnqueueBuffers.
        #
        # Looping sources will never fully process buffers because they will be set to
        # play again for when the source loops.
        AL_BUFFERS_PROCESSED = 0x1016

        # Source reference distance.
        # Type:    ALfloat
        # Range:   [0.0 - ]
        # Default: 1.0
        #
        # The distance in units that no distance attenuation occurs.
        #
        # At 0.0, no distance attenuation occurs with non-linear attenuation models.
        AL_REFERENCE_DISTANCE = 0x1020

        # Source rolloff factor.
        # Type:    ALfloat
        # Range:   [0.0 - ]
        # Default: 1.0
        #
        # Multiplier to exaggerate or diminish distance attenuation.
        #
        # At 0.0, no distance attenuation ever occurs.
        AL_ROLLOFF_FACTOR = 0x1021

        # Outer cone gain.
        # Type:    ALfloat
        # Range:   [0.0 - 1.0]
        # Default: 0.0
        #
        # The gain attenuation applied when the listener is outside of the source's
        # outer cone angle.
        AL_CONE_OUTER_GAIN = 0x1022

        # Source maximum distance.
        # Type:    ALfloat
        # Range:   [0.0 - ]
        # Default: FLT_MAX
        #
        # The distance above which the source is not attenuated any further with a
        # clamped distance model, or where attenuation reaches 0.0 gain for linear
        # distance models with a default rolloff factor.
        AL_MAX_DISTANCE = 0x1023

        #  Source buffer offset, in seconds
        AL_SEC_OFFSET = 0x1024
        #  Source buffer offset, in sample frames
        AL_SAMPLE_OFFSET = 0x1025
        #  Source buffer offset, in bytes
        AL_BYTE_OFFSET = 0x1026

        # Source type (query only).
        # Type:  ALenum
        # Range: [AL_STATIC, AL_STREAMING, AL_UNDETERMINED]
        #
        # A Source is Static if a Buffer has been attached using AL_BUFFER.
        #
        # A Source is Streaming if one or more Buffers have been attached using
        # alSourceQueueBuffers.
        #
        # A Source is Undetermined when it has the NULL buffer attached using
        # AL_BUFFER.
        AL_SOURCE_TYPE = 0x1027

        #  Source type values.
        AL_STATIC = 0x1028
        AL_STREAMING = 0x1029
        AL_UNDETERMINED = 0x1030

        #  Unsigned 8-bit mono buffer format.
        AL_FORMAT_MONO8 = 0x1100
        #  Signed 16-bit mono buffer format.
        AL_FORMAT_MONO16 = 0x1101
        #  Unsigned 8-bit stereo buffer format.
        AL_FORMAT_STEREO8 = 0x1102
        #  Signed 16-bit stereo buffer format.
        AL_FORMAT_STEREO16 = 0x1103

        #  Buffer frequency/sample rate (query only).
        AL_FREQUENCY = 0x2001
        #  Buffer bits per sample (query only).
        AL_BITS = 0x2002
        #  Buffer channel count (query only).
        AL_CHANNELS = 0x2003
        #  Buffer data size in bytes (query only).
        AL_SIZE = 0x2004

        #  Buffer state. Not for public use.
        AL_UNUSED = 0x2010
        AL_PENDING = 0x2011
        AL_PROCESSED = 0x2012


        #  No error.
        AL_NO_ERROR = 0

        #  Invalid name (ID) passed to an AL call.
        AL_INVALID_NAME = 0xA001

        #  Invalid enumeration passed to AL call.
        AL_INVALID_ENUM = 0xA002

        #  Invalid value passed to AL call.
        AL_INVALID_VALUE = 0xA003

        #  Illegal AL call.
        AL_INVALID_OPERATION = 0xA004

        #  Not enough memory to execute the AL call.
        AL_OUT_OF_MEMORY = 0xA005


        #  Context string: Vendor name.
        AL_VENDOR = 0xB001
        #  Context string: Version.
        AL_VERSION = 0xB002
        #  Context string: Renderer name.
        AL_RENDERER = 0xB003
        #  Context string: Space-separated extension list.
        AL_EXTENSIONS = 0xB004

        # Doppler scale.
        # Type:    ALfloat
        # Range:   [0.0 - ]
        # Default: 1.0
        #
        # Scale for source and listener velocities.
        AL_DOPPLER_FACTOR = 0xC000

        # Doppler velocity (deprecated).
        #
        # A multiplier applied to the Speed of Sound.
        AL_DOPPLER_VELOCITY = 0xC001

        # Speed of Sound, in units per second.
        # Type:    ALfloat
        # Range:   [0.0001 - ]
        # Default: 343.3
        #
        # The speed at which sound waves are assumed to travel, when calculating the
        # doppler effect from source and listener velocities.
        AL_SPEED_OF_SOUND = 0xC003

        # Distance attenuation model.
        # Type:    ALenum
        # Range:   [AL_NONE, AL_INVERSE_DISTANCE, AL_INVERSE_DISTANCE_CLAMPED,
        #           AL_LINEAR_DISTANCE, AL_LINEAR_DISTANCE_CLAMPED,
        #           AL_EXPONENT_DISTANCE, AL_EXPONENT_DISTANCE_CLAMPED]
        # Default: AL_INVERSE_DISTANCE_CLAMPED
        #
        # The model by which sources attenuate with distance.
        #
        # None     - No distance attenuation.
        # Inverse  - Doubling the distance halves the source gain.
        # Linear   - Linear gain scaling between the reference and max distances.
        # Exponent - Exponential gain dropoff.
        #
        # Clamped variations work like the non-clamped counterparts, except the
        # distance calculated is clamped between the reference and max distances.
        AL_DISTANCE_MODEL = 0xD000

        #  Distance model values.
        AL_INVERSE_DISTANCE = 0xD001
        AL_INVERSE_DISTANCE_CLAMPED = 0xD002
        AL_LINEAR_DISTANCE = 0xD003
        AL_LINEAR_DISTANCE_CLAMPED = 0xD004
        AL_EXPONENT_DISTANCE = 0xD005
        AL_EXPONENT_DISTANCE_CLAMPED = 0xD006

    #  Renderer State management.
    void alEnable(ALenum capability)
    void alDisable(ALenum capability)
    ALboolean alIsEnabled(ALenum capability)

    #  Context state setting.
    void alDopplerFactor(ALfloat value)
    void alDopplerVelocity(ALfloat value)
    void alSpeedOfSound(ALfloat value)
    void alDistanceModel(ALenum distanceModel)

    #  Context state retrieval.
    const ALchar* alGetString(ALenum param)
    void alGetBooleanv(ALenum param, ALboolean *values)
    void alGetIntegerv(ALenum param, ALint *values)
    void alGetFloatv(ALenum param, ALfloat *values)
    void alGetDoublev(ALenum param, ALdouble *values)
    ALboolean alGetBoolean(ALenum param)
    ALint alGetInteger(ALenum param)
    ALfloat alGetFloat(ALenum param)
    ALdouble alGetDouble(ALenum param)

    # Obtain the first error generated in the AL context since the last call to
    # this function.
    ALenum alGetError()

    #  Query for the presence of an extension on the AL context.
    ALboolean alIsExtensionPresent(const ALchar *extname)
    # Retrieve the address of a function. The returned function may be context-
    # specific.
    void* alGetProcAddress(const ALchar *fname)
    # Retrieve the value of an enum. The returned value may be context-specific.
    ALenum alGetEnumValue(const ALchar *ename)


    #  Set listener parameters.
    void alListenerf(ALenum param, ALfloat value)
    void alListener3f(ALenum param, ALfloat value1, ALfloat value2, ALfloat value3)
    void alListenerfv(ALenum param, const ALfloat *values)
    void alListeneri(ALenum param, ALint value)
    void alListener3i(ALenum param, ALint value1, ALint value2, ALint value3)
    void alListeneriv(ALenum param, const ALint *values)

    #  Get listener parameters.
    void alGetListenerf(ALenum param, ALfloat *value)
    void alGetListener3f(ALenum param, ALfloat *value1, ALfloat *value2, ALfloat *value3)
    void alGetListenerfv(ALenum param, ALfloat *values)
    void alGetListeneri(ALenum param, ALint *value)
    void alGetListener3i(ALenum param, ALint *value1, ALint *value2, ALint *value3)
    void alGetListeneriv(ALenum param, ALint *values)


    #  Create source objects.
    void alGenSources(ALsizei n, ALuint *sources)
    #  Delete source objects.
    void alDeleteSources(ALsizei n, const ALuint *sources)
    #  Verify an ID is for a valid source.
    ALboolean alIsSource(ALuint source)

    #  Set source parameters.
    void alSourcef(ALuint source, ALenum param, ALfloat value)
    void alSource3f(ALuint source, ALenum param, ALfloat value1, ALfloat value2, ALfloat value3)
    void alSourcefv(ALuint source, ALenum param, const ALfloat *values)
    void alSourcei(ALuint source, ALenum param, ALint value)
    void alSource3i(ALuint source, ALenum param, ALint value1, ALint value2, ALint value3)
    void alSourceiv(ALuint source, ALenum param, const ALint *values)

    #  Get source parameters.
    void alGetSourcef(ALuint source, ALenum param, ALfloat *value)
    void alGetSource3f(ALuint source, ALenum param, ALfloat *value1, ALfloat *value2, ALfloat *value3)
    void alGetSourcefv(ALuint source, ALenum param, ALfloat *values)
    void alGetSourcei(ALuint source,  ALenum param, ALint *value)
    void alGetSource3i(ALuint source, ALenum param, ALint *value1, ALint *value2, ALint *value3)
    void alGetSourceiv(ALuint source,  ALenum param, ALint *values)


    #  Play, restart, or resume a source, setting its state to AL_PLAYING.
    void alSourcePlay(ALuint source)
    #  Stop a source, setting its state to AL_STOPPED if playing or paused.
    void alSourceStop(ALuint source)
    #  Rewind a source, setting its state to AL_INITIAL.
    void alSourceRewind(ALuint source)
    #  Pause a source, setting its state to AL_PAUSED if playing.
    void alSourcePause(ALuint source)

    #  Play, restart, or resume a list of sources atomically.
    void alSourcePlayv(ALsizei n, const ALuint *sources)
    #  Stop a list of sources atomically.
    void alSourceStopv(ALsizei n, const ALuint *sources)
    #  Rewind a list of sources atomically.
    void alSourceRewindv(ALsizei n, const ALuint *sources)
    #  Pause a list of sources atomically.
    void alSourcePausev(ALsizei n, const ALuint *sources)

    #  Queue buffers onto a source
    void alSourceQueueBuffers(ALuint source, ALsizei nb, const ALuint *buffers)
    #  Unqueue processed buffers from a source
    void alSourceUnqueueBuffers(ALuint source, ALsizei nb, ALuint *buffers)


    #  Create buffer objects
    void alGenBuffers(ALsizei n, ALuint *buffers)
    #  Delete buffer objects
    void alDeleteBuffers(ALsizei n, const ALuint *buffers)
    #  Verify an ID is a valid buffer (including the NULL buffer)
    ALboolean alIsBuffer(ALuint buffer)

    # Copies data into the buffer, interpreting it using the specified format and
    # samplerate.
    void alBufferData(ALuint buffer, ALenum format, const ALvoid *data, ALsizei size, ALsizei samplerate)

    #  Set buffer parameters.
    void alBufferf(ALuint buffer, ALenum param, ALfloat value)
    void alBuffer3f(ALuint buffer, ALenum param, ALfloat value1, ALfloat value2, ALfloat value3)
    void alBufferfv(ALuint buffer, ALenum param, const ALfloat *values)
    void alBufferi(ALuint buffer, ALenum param, ALint value)
    void alBuffer3i(ALuint buffer, ALenum param, ALint value1, ALint value2, ALint value3)
    void alBufferiv(ALuint buffer, ALenum param, const ALint *values)

    #  Get buffer parameters.
    void alGetBufferf(ALuint buffer, ALenum param, ALfloat *value)
    void alGetBuffer3f(ALuint buffer, ALenum param, ALfloat *value1, ALfloat *value2, ALfloat *value3)
    void alGetBufferfv(ALuint buffer, ALenum param, ALfloat *values)
    void alGetBufferi(ALuint buffer, ALenum param, ALint *value)
    void alGetBuffer3i(ALuint buffer, ALenum param, ALint *value1, ALint *value2, ALint *value3)
    void alGetBufferiv(ALuint buffer, ALenum param, ALint *values)



# cython: language_level=3

from .context cimport Context
from .util cimport V3f
from . cimport al

cdef class Listener:
    def __cinit__(self, Context ctx):
        self.context = ctx

    @property
    def position(self):
        cdef float x, y, z
        al.alGetListener3f(al.AL_POSITION, &x, &y, &z)
        return V3f(x, y, z)

    @position.setter
    def position(self, pos):
        al.alListener3f(al.AL_POSITION, pos[0], pos[1], pos[2])

    @property
    def velocity(self):
        cdef float x, y, z
        al.alGetListener3f(al.AL_VELOCITY, &x, &y, &z)
        return V3f(x, y, z)

    @velocity.setter
    def velocity(self, val):
        al.alListener3f(al.AL_VELOCITY, val[0], val[1], val[2])

    @property
    def gain(self):
        cdef float val
        al.alGetListenerf(al.AL_GAIN, &val)
        return val

    @gain.setter
    def gain(self, val):
        al.alListenerf(al.AL_GAIN, val)

    @property
    def orientation(self):
        cdef float[6] data
        al.alGetListenerfv(al.AL_ORIENTATION, data)
        return (V3f(data[0], data[1], data[2]), V3f(data[3], data[4], data[5]))

    @orientation.setter
    def orientation(self, val):
        cdef float[6] data
        if len(val) == 2:
            data = [val[0][0], val[0][1], val[0][2], val[1][0], val[1][1], val[1][2]]
        elif len(val) >= 6:
            data = [val[0], val[1], val[2], val[3], val[4], val[5]]
        else:
            raise ValueError("Invalid length for 'orientation'")
        al.alListenerfv(al.AL_ORIENTATION, data)



import cyal, math, time, wave

# Open an OpenAL device and create a Context from it
# Unknown keyword arguments such as hrtf_soft below are passed directly to OpenAL. The hrtf_soft=1 argument enables
# OpenAL Soft's Head Related Transfer Function processing
ctx = cyal.Context(cyal.Device(), make_current=True, hrtf_soft=1)

# Create two buffer objects to load the left and right channel audio into
# We use two separate buffers so they can be attached to two separate sources, and moved around the world separately
# Context.gen_buffers(n) may be faster than Context.gen_buffer() when generating Buffer objects in bulk
left_buf, right_buf = ctx.gen_buffers(2)

# Load wave data into the buffers
# This uses Python's built-in wave module to load the audio from wave files in the examples directory
# Music by Mudb0y <https://soundcloud.com/mudb0y>
with wave.open("beat2-left.wav", "r") as w:
    data = w.readframes(w.getnframes())
    left_buf.set_data(data, sample_rate=w.getframerate(), format=cyal.BufferFormat.MONO16)

with wave.open("beat2-right.wav", "r") as w:
    data = w.readframes(w.getnframes())
    right_buf.set_data(data, sample_rate=w.getframerate(), format=cyal.BufferFormat.MONO16)

# Generate Source objects
# You can pass properties to the Context.gen_*() functions, they will be set on the returned objects
# In this case, ctx.gen_sources(2 ...
left_src, right_src = ctx.gen_sources(2, looping=True)
# ... is the same as ...
# left_src, right_src = ctx.gen_sources(2)
# left_src.looping = True
# right_src.looping = True

# Assign the buffers to the sources
left_src.buffer = left_buf
right_src.buffer = right_buf

# Set source positions
# OpenAL's coordinates are left-handed: +x = right, +y = up, +z = backwards
# This sets the two sources to be in front of the listener at the start, separated as if they were stereo speakers in a room
left_src.position = [-1.5, 0, -1]
right_src.position = [1.5, 0, -1]

# Set listener properties (these are the defaults)
ctx.listener.position = [0, 0, 0]
# The first 3 coordinates are the facing direction, the last 3 are the up direction
ctx.listener.orientation = [0, 0, -1, 0, 1, 0]

# Start playback of both sources
# You can call Source.play() on a single source, but Context.play_sources() ensures that the sources start playback
# atomically, at *exactly* the same time
ctx.play_sources(left_src, right_src)

# Spin the listener around in a circle
angle = 0.0
while True:
    # Calculate the new coordinates (Y remains constant)
    new_x = math.sin(angle)
    new_z = math.cos(angle)
    # Update the listener orientation
    ctx.listener.orientation = [new_x, 0, new_z, 0, 1, 0]
    angle += 0.01
    time.sleep(0.02)



# cython: language_level=3

from . cimport al, alc

cdef class Device:
    cdef object __weakref__
    cdef alc.ALCdevice* _device
    cdef al.ALvoid* (*get_al_proc_address)(const al.ALchar*)
    cdef void (*pause_soft)(alc.ALCdevice*)
    cdef void (*resume_soft)(alc.ALCdevice*)
    cdef alc.ALCboolean (*alc_reset_device_soft)(alc.ALCdevice* device, const alc.ALCint* attribs)
    cdef alc.ALCboolean (*alc_reopen_device_soft)(alc.ALCdevice* device, const alc.ALCchar* name, const alc.ALCint* attribs)

    # Flags for extension emulation
    cdef bint emulate_enumerate_all

    cdef alc.ALCenum get_enum_value(self, const alc.ALCchar* name)
    cdef inline alc.ALCvoid* get_alc_proc_address(self, const alc.ALCchar *funcname):
        return alc.alcGetProcAddress(self._device, funcname)

    cpdef list get_supported_extensions(self)



# cython: language_level=3

from cpython cimport array
import array
from contextlib import contextmanager

from .buffer cimport Buffer
from .device cimport Device
from .exceptions cimport UnsupportedExtensionError, check_al_error, check_alc_error
from .listener cimport Listener
from .source cimport Source
from . cimport al, alc

cdef array.array ids_template = array.array('I')

cdef class Context:
    def __cinit__(self, Device dev not None, *,
            make_current=False,
            emulate_deferred_updates=True,
            emulate_direct_channels=True,
            emulate_direct_channels_remix=True,
            emulate_source_spatialize=True,
            emulate_disconnect=True,
            emulate_buffer_length_query=True,
            **kwargs):
        self.device = dev
        cdef ContextAttrs attrs = ContextAttrs.from_kwargs(dev, **kwargs)
        self._ctx  =alc.alcCreateContext(dev._device, &attrs._attrs[0])
        if self._ctx is NULL:
            check_alc_error(dev._device)
        self.listener = Listener(self)

    # Flags for extension emulation
        self.emulate_direct_channels = emulate_direct_channels
        self.emulate_direct_channels_remix = emulate_direct_channels_remix
        self.emulate_source_spatialize = emulate_source_spatialize
        self.emulate_disconnect = emulate_disconnect
        self.emulate_buffer_length_query = emulate_buffer_length_query

        # Make the context current here, as checking for extensions requires it, and alGetProcAddress() may return context-specific functions
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(self._ctx)

        # Extension enum values
        self.alc_connected = alc.alcGetEnumValue(self.device._device, b"ALC_CONNECTED");
        if al.alIsExtensionPresent(b"AL_SOFT_BUFFER_LENGTH_QUERY") == al.AL_TRUE:
            self.al_byte_length_soft = al.alGetEnumValue(b"AL_BYTE_LENGTH_SOFT")
            self.al_sample_length_soft = al.alGetEnumValue(b"AL_SAMPLE_LENGTH_SOFT")
            self.al_sec_length_soft = al.alGetEnumValue(b"AL_SEC_LENGTH_SOFT")
        else:
            self.al_byte_length_soft = al.AL_NONE
            self.al_sample_length_soft = al.AL_NONE
            self.al_sec_length_soft = al.AL_NONE

        # AL_SOFT_deferred_updates extension functions
        if al.alIsExtensionPresent("AL_SOFT_DEFERRED_UPDATES") == al.AL_TRUE:
            self.al_defer_updates_soft = <void (*)()>dev.get_al_proc_address("alDeferUpdatesSOFT")
            self.al_process_updates_soft = <void (*)()>dev.get_al_proc_address("alProcessUpdatesSOFT")
        elif emulate_deferred_updates:
            self.al_defer_updates_soft = wrap_defer_updates
            self.al_process_updates_soft = wrap_process_updates
        else:
            self.al_defer_updates_soft = no_defer_updates
            self.al_process_updates_soft = no_process_updates

        if not make_current:
            # Restore the previous context
            alc.alcMakeContextCurrent(prev_ctx)

    def __dealloc__(self):
        if self._ctx:
            if self._ctx == alc.alcGetCurrentContext():
                alc.alcMakeContextCurrent(NULL)
            alc.alcDestroyContext(self._ctx)

    def get_attrs(self):
        return ContextAttrs.from_device(self.device)

    @property
    def supported_extensions(self):
        cdef const al.ALchar* exts = al.alGetString(al.AL_EXTENSIONS)
        return exts.decode("utf8").split(' ') if exts is not NULL else []

    def is_extension_present(self, ext_name):
        return al.alIsExtensionPresent(ext_name.encode("utf8")) == al.AL_TRUE

    def make_current(self):
        alc.alcMakeContextCurrent(self._ctx)

    @property
    def is_connected(self):
        cdef al.ALint val
        if self.alc_connected != al.AL_NONE:
            alc.alcGetIntegerv(self.device._device, self.alc_connected, 1, &val)
            check_alc_error(self.device._device)
            return val != al.AL_FALSE
        elif self.emulate_disconnect:
            return True
        else:
            raise UnsupportedExtensionError("ALC_EXT_DISCONNECT")

    @property
    def is_current(self):
        return self._ctx == alc.alcGetCurrentContext()

    @contextmanager
    def as_current(self):
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(self._ctx)
        try:
            yield self
        finally:
            alc.alcMakeContextCurrent(prev_ctx)

    def process(self):
        alc.alcProcessContext(self._ctx)

    def suspend(self):
        alc.alcSuspendContext(self._ctx)

    @contextmanager
    def as_processing(self):
        alc.alcProcessContext(self._ctx)
        try:
            yield self
        finally:
            alc.alcSuspendContext(self._ctx)

    @contextmanager
    def as_suspended(self):
        alc.alcSuspendContext(self._ctx)
        try:
            yield self
        finally:
            alc.alcProcessContext(self._ctx)

    def defer_updates(self):
        self.al_defer_updates_soft()

    def process_updates(self):
        self.al_process_updates_soft()

    @contextmanager
    def batch(self):
        self.al_defer_updates_soft()
        try:
            yield self
        finally:
            self.al_process_updates_soft()

    def gen_buffer(self):
        cdef al.ALuint id
        al.alGenBuffers(1, &id)
        check_al_error()
        return Buffer.from_id(self, id)

    def gen_buffers(self, n):
        cdef al.ALuint[:] ids = array.clone(ids_template, n, zero=False)
        al.alGenBuffers(n, &ids[0])
        check_al_error()
        return [Buffer.from_id(self, id) for id in ids]

    def gen_source(self, **kwargs):
        cdef al.ALuint id
        al.alGenSources(1, &id)
        check_al_error()
        cdef Source src = Source.from_id(self, id)
        for k, v in kwargs.items(): setattr(src, k, v)
        return src

    def gen_sources(self, n, **kwargs):
        cdef al.ALuint[:] ids = array.clone(ids_template, n, zero=False)
        al.alGenSources(n, &ids[0])
        check_al_error()
        cdef list srcs = [Source.from_id(self, id) for id in ids]
        for src in srcs:
            for k, v in kwargs.items(): setattr(src, k, v)
        return srcs

    def play_sources(self, *srcs):
        cdef al.ALuint[:] ids = array.clone(ids_template, len(srcs), zero=False)
        cdef Py_ssize_t i
        for i, s in enumerate(srcs): ids[i] = s.id
        al.alSourcePlayv(ids.size, &ids[0])
        check_al_error()

    def stop_sources(self, *srcs):
        cdef al.ALuint[:] ids = array.clone(ids_template, len(srcs), zero=False)
        cdef Py_ssize_t i
        for i, s in enumerate(srcs): ids[i] = s.id
        al.alSourceStopv(ids.size, &ids[0])
        check_al_error()

    def rewind_sources(self, *srcs):
        cdef al.ALuint[:] ids = array.clone(ids_template, len(srcs), zero=False)
        cdef Py_ssize_t i
        for i, s in enumerate(srcs): ids[i] = s.id
        al.alSourceRewindv(ids.size, &ids[0])
        check_al_error()

    def pause_sources(self, *srcs):
        cdef al.ALuint[:] ids = array.clone(ids_template, len(srcs), zero=False)
        cdef Py_ssize_t i
        for i, s in enumerate(srcs): ids[i] = s.id
        al.alSourcePausev(ids.size, &ids[0])
        check_al_error()

cdef array.array attrs_template = array.array('i')

cdef class ContextAttrs:
    def __cinit__(self, Device dev, array.array attrs):
        self._attrs = attrs
        self._dev = dev

    @staticmethod
    def from_kwargs(Device dev, **kwargs):
        cdef:
            Py_ssize_t length = len(kwargs) * 2 + 1
            array.array attrs_array = array.clone(attrs_template, length, zero=False)
            alc.ALCint[:] attrs = attrs_array
            bytes enum_name
            alc.ALCenum enum_val
            Py_ssize_t idx = 0
        for k, v in kwargs.items():
            enum_name = b"ALC_" + k.upper().encode("utf-8")
            enum_val = alc.alcGetEnumValue(dev._device, <const alc.ALCchar *>enum_name)
            if enum_val == al.AL_NONE:
                raise TypeError(f"'{k} is an invalid keyword argument for ContextAttrs")
            attrs[idx] = enum_val
            attrs[idx+1] = <alc.ALCint>v
            idx += 2
        attrs[idx] = 0
        return ContextAttrs(dev, attrs_array)

    @staticmethod
    def from_device(Device dev):
        cdef alc.ALCint length
        alc.alcGetIntegerv(dev._device, alc.ALC_ATTRIBUTES_SIZE, 1, &length)
        cdef array.array attrs = array.clone(attrs_template, length + 1, zero=False)
        alc.alcGetIntegerv(dev._device, alc.ALC_ALL_ATTRIBUTES, length, <alc.ALCint *>attrs.data.as_voidptr)
        attrs[length] = 0
        return ContextAttrs(dev, attrs)

    def __getattr__(self, attr):
        cdef:
            bytes enum_name = b"ALC_" + attr.upper().encode("utf-8")
            alc.ALCenum enum_val = alc.alcGetEnumValue(self._dev._device, <const alc.ALCchar *>enum_name)
        if enum_val == al.AL_NONE:
            raise AttributeError(f"ContextAttrs object has no attribute '{attr}'")
        cdef Py_ssize_t i
        for i in range(0, self._attrs.size, 2):
            if self._attrs[i] == enum_val:
                return self._attrs[i+1]
        # Not found
        raise AttributeError(f"ContextAttrs object has no attribute '{attr}'")

cdef void wrap_defer_updates():
    alc.alcSuspendContext(alc.alcGetCurrentContext())

cdef void wrap_process_updates():
    alc.alcProcessContext(alc.alcGetCurrentContext())

cdef void no_defer_updates(): pass
cdef void no_process_updates(): pass



# cython: language_level=3

from cpython cimport array
import array
from collections.abc import Sequence

from . cimport al, alc
from .context cimport Context
from .device cimport Device
from .exceptions cimport check_alc_error, check_al_error
from .util cimport get_al_enum, V3f

cdef array.array ids_template = array.array('I')

cdef class EfxExtension:
    def __cinit__(self, Context ctx not None):
        if alc.alcIsExtensionPresent(ctx.device._device, "ALC_EXT_EFX") == al.AL_FALSE:
            raise RuntimeError("Unsupported extension: ALC_EXT_EFX")
        self.context = ctx

        # Get pointers to all the needed functions
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(ctx._ctx)

        self.alGenEffects = <void (*)(al.ALsizei n, al.ALuint *effects)>al.alGetProcAddress("alGenEffects")
        self.alDeleteEffects = <void (*)(al.ALsizei n, const al.ALuint *effects)>al.alGetProcAddress("alDeleteEffects")
        self.alEffecti = <void (*)(al.ALuint effect, al.ALenum param, al.ALint iValue)>al.alGetProcAddress("alEffecti")
        self.alEffectiv = <void (*)(al.ALuint effect, al.ALenum param, const al.ALint *piValues)>al.alGetProcAddress("alEffectiv")
        self.alEffectf = <void (*)(al.ALuint effect, al.ALenum param, al.ALfloat flValue)>al.alGetProcAddress("alEffectf")
        self.alEffectfv = <void (*)(al.ALuint effect, al.ALenum param, const al.ALfloat *pflValues)>al.alGetProcAddress("alEffectfv")
        self.alGetEffecti = <void (*)(al.ALuint effect, al.ALenum param, al.ALint *piValue)>al.alGetProcAddress("alGetEffecti")
        self.alGetEffectiv = <void (*)(al.ALuint effect, al.ALenum param, al.ALint *piValues)>al.alGetProcAddress("alGetEffectiv")
        self.alGetEffectf = <void (*)(al.ALuint effect, al.ALenum param, al.ALfloat *pflValue)>al.alGetProcAddress("alGetEffectf")
        self.alGetEffectfv = <void (*)(al.ALuint effect, al.ALenum param, al.ALfloat *pflValues)>al.alGetProcAddress("alGetEffectfv")

        self.alGenFilters = <void (*)(al.ALsizei n, al.ALuint *filters)>al.alGetProcAddress("alGenFilters")
        self.alDeleteFilters = <void (*)(al.ALsizei n, const al.ALuint *filters)>al.alGetProcAddress("alDeleteFilters")
        self.alFilteri = <void (*)(al.ALuint filter, al.ALenum param, al.ALint iValue)>al.alGetProcAddress("alFilteri")
        self.alFilteriv = <void (*)(al.ALuint filter, al.ALenum param, const al.ALint *piValues)>al.alGetProcAddress("alFilteriv")
        self.alFilterf = <void (*)(al.ALuint filter, al.ALenum param, al.ALfloat flValue)>al.alGetProcAddress("alFilterf")
        self.alFilterfv = <void (*)(al.ALuint filter, al.ALenum param, const al.ALfloat *pflValues)>al.alGetProcAddress("alFilterfv")
        self.alGetFilteri = <void (*)(al.ALuint filter, al.ALenum param, al.ALint *piValue)>al.alGetProcAddress("alGetFilteri")
        self.alGetFilteriv = <void (*)(al.ALuint filter, al.ALenum param, al.ALint *piValues)>al.alGetProcAddress("alGetFilteriv")
        self.alGetFilterf = <void (*)(al.ALuint filter, al.ALenum param, al.ALfloat *pflValue)>al.alGetProcAddress("alGetFilterf")
        self.alGetFilterfv = <void (*)(al.ALuint filter, al.ALenum param, al.ALfloat *pflValues)>al.alGetProcAddress("alGetFilterfv")

        self.alGenAuxiliaryEffectSlots = <void (*)(al.ALsizei n, al.ALuint *effectslots)>al.alGetProcAddress("alGenAuxiliaryEffectSlots")
        self.alDeleteAuxiliaryEffectSlots = <void (*)(al.ALsizei n, const al.ALuint *effectslots)>al.alGetProcAddress("alDeleteAuxiliaryEffectSlots")
        self.alAuxiliaryEffectSloti = <void (*)(al.ALuint effectslot, al.ALenum param, al.ALint iValue)>al.alGetProcAddress("alAuxiliaryEffectSloti")
        self.alAuxiliaryEffectSlotiv = <void (*)(al.ALuint effectslot, al.ALenum param, const al.ALint *piValues)>al.alGetProcAddress("alAuxiliaryEffectSlotiv")
        self.alAuxiliaryEffectSlotf = <void (*)(al.ALuint effectslot, al.ALenum param, al.ALfloat flValue)>al.alGetProcAddress("alAuxiliaryEffectSlotf")
        self.alAuxiliaryEffectSlotfv = <void (*)(al.ALuint effectslot, al.ALenum param, const al.ALfloat *pflValues)>al.alGetProcAddress("alAuxiliaryEffectSlotfv")
        self.alGetAuxiliaryEffectSloti = <void (*)(al.ALuint effectslot, al.ALenum param, al.ALint *piValue)>al.alGetProcAddress("alGetAuxiliaryEffectSloti")
        self.alGetAuxiliaryEffectSlotiv = <void (*)(al.ALuint effectslot, al.ALenum param, al.ALint *piValues)>al.alGetProcAddress("alGetAuxiliaryEffectSlotiv")
        self.alGetAuxiliaryEffectSlotf = <void (*)(al.ALuint effectslot, al.ALenum param, al.ALfloat *pflValue)>al.alGetProcAddress("alGetAuxiliaryEffectSlotf")
        self.alGetAuxiliaryEffectSlotfv = <void (*)(al.ALuint effectslot, al.ALenum param, al.ALfloat *pflValues)>al.alGetProcAddress("alGetAuxiliaryEffectSlotfv")

        self.AL_METERS_PER_UNIT = alc.alcGetEnumValue(ctx.device._device, b"AL_METERS_PER_UNIT")
        self.alc_max_auxiliary_sends = alc.alcGetEnumValue(ctx.device._device, b"ALC_MAX_AUXILIARY_SENDS")
        self.AL_EFFECT_TYPE = alc.alcGetEnumValue(ctx.device._device, b"AL_EFFECT_TYPE")
        self.AL_FILTER_TYPE = alc.alcGetEnumValue(ctx.device._device, b"AL_FILTER_TYPE")

        # Restore the old context (if any)
        alc.alcMakeContextCurrent(prev_ctx)

    @property
    def version(self):
        cdef alc.ALCenum ver_maj = self.context.device.get_enum_value("ALC_EFX_MAJOR_VERSION")
        cdef alc.ALCenum ver_min = self.context.device.get_enum_value("ALC_EFX_MINOR_VERSION")
        cdef alc.ALCint maj, min
        alc.alcGetIntegerv(self.context.device._device, ver_maj, 1, &maj)
        alc.alcGetIntegerv(self.context.device._device, ver_min, 1, &min)
        check_alc_error(self.context.device._device)
        return (maj, min)

    @property
    def meters_per_unit(self):
        cdef al.ALfloat mpu
        al.alGetListenerf(self.AL_METERS_PER_UNIT, &mpu)
        return mpu

    @meters_per_unit.setter
    def meters_per_unit(self, val):
        al.alListenerf(self.AL_METERS_PER_UNIT, val)
        check_al_error()

    @property
    def max_auxiliary_sends(self):
        cdef al.ALint val
        alc.alcGetIntegerv(self.context.device._device, self.alc_max_auxiliary_sends, 1, &val)
        check_alc_error(self.context.device._device)
        return val

    def gen_auxiliary_effect_slot(self, **kwargs):
        cdef al.ALuint id
        self.alGenAuxiliaryEffectSlots(1, &id)
        check_al_error()
        cdef AuxiliaryEffectSlot slot = AuxiliaryEffectSlot.from_id(self, id)
        for k, v in kwargs.items(): setattr(slot, k, v)
        return slot

    def gen_auxiliary_effect_slots(self, n, **kwargs):
        cdef al.ALuint[:] ids = array.clone(ids_template, n, zero=False)
        self.alGenAuxiliaryEffectSlots(n, &ids[0])
        check_al_error()
        cdef list slots = [AuxiliaryEffectSlot.from_id(self, id) for id in ids]
        for slot in slots:
            for k, v in kwargs.items(): setattr(slot, k, v)
        return slots

    def gen_effect(self, type cls=Effect, **kwargs):
        if not issubclass(cls, Effect):
            raise TypeError("cls argument must be a subclass of cyal.efx.Effect")
        cdef al.ALuint id
        self.alGenEffects(1, &id)
        check_al_error()
        return make_effect(cls, self, id, kwargs)

    def gen_effects(self, n, cls=Effect, **kwargs):
        if not issubclass(cls, Effect):
            raise TypeError("cls argument must be a subclass of cyal.efx.Effect")
        cdef al.ALuint[:] ids = array.clone(ids_template, n, zero=False)
        self.alGenEffects(n, &ids[0])
        check_al_error()
        return [make_effect(cls, self, id, kwargs) for id in ids]

    def gen_filter(self, type cls=Filter, **kwargs):
        if not issubclass(cls, Filter):
            raise TypeError("cls argument must be a subclass of cyal.efx.Filter")
        cdef al.ALuint id
        self.alGenFilters(1, &id)
        check_al_error()
        return make_filter(cls, self, id, kwargs)

    def gen_filters(self, n, cls=Filter, **kwargs):
        if not issubclass(cls, Filter):
            raise TypeError("cls argument must be a subclass of cyal.efx.Filter")
        cdef al.ALuint[:] ids = array.clone(ids_template, n, zero=False)
        self.alGenFilters(n, &ids[0])
        check_al_error()
        return [make_filter(cls, self, id, kwargs) for id in ids]

cdef class AuxiliaryEffectSlot:
    def __cinit__(self):
        pass

    def __init__(self):
        raise TypeError("This class cannot be instantiated directly.")

    @staticmethod
    cdef AuxiliaryEffectSlot from_id(EfxExtension efx, al.ALuint id):
        cdef AuxiliaryEffectSlot slot = AuxiliaryEffectSlot.__new__(AuxiliaryEffectSlot)
        slot.efx = efx
        slot.id = id
        return slot

    def __dealloc__(self):
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(self.efx.context._ctx)
        self.efx.alDeleteAuxiliaryEffectSlots(1, &self.id)
        alc.alcMakeContextCurrent(prev_ctx)

cdef class Effect:
    def __cinit__(self):
        self._type = "null"

    def __init__(self):
        raise TypeError("This class cannot be instantiated directly.")

    cdef void init_with_id(self, EfxExtension efx, al.ALuint id):
        self.efx = efx
        self.id = id

    def __dealloc__(self):
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(self.efx.context._ctx)
        self.efx.alDeleteEffects(1, &self.id)
        alc.alcMakeContextCurrent(prev_ctx)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, val):
        cdef str val_upper = val.upper()
        cdef enum_val = al.alGetEnumValue(b"AL_EFFECT_" + val_upper.encode("utf8"))
        if enum_val == al.AL_NONE:
            raise RuntimeError(f"Could not get enum value for AL_EFFECT_{val_upper}")
        self.efx.alEffecti(self.id, self.efx.AL_EFFECT_TYPE, enum_val)
        check_al_error()
        self._type = val

    def get_int(self, prop):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALint val
        self.efx.alGetEffecti(self.id, e_val, &val)
        check_al_error()
        return val

    def get_float(self, prop):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALfloat val
        self.efx.alGetEffectf(self.id, e_val, &val)
        check_al_error()
        return val

    def get_v3f(self, prop):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALfloat[3] val
        self.efx.alGetEffectfv(self.id, e_val, val)
        check_al_error()
        return V3f(val[0], val[1], val[2])

    def set(self, prop, val):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALfloat[3] data
        if isinstance(val, int):
            self.efx.alEffecti(self.id, e_val, val)
        elif isinstance(val, float):
            self.efx.alEffectf(self.id, e_val, val)
        elif (isinstance(val, Sequence) and len(val) == 3) or isinstance(val, V3f):
            data = [val[0], val[1], val[2]]
            self.efx.alEffectfv(self.id, e_val, data)
        else:
            raise TypeError("Cannot convert to OpenAL type")
        check_al_error()

cdef class Filter:
    def __cinit__(self):
        self._type = "null"

    def __init__(self):
        raise TypeError("This class cannot be instantiated directly.")

    cdef void init_with_id(self, EfxExtension efx, al.ALuint id):
        self.efx = efx
        self.id = id

    def __dealloc__(self):
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(self.efx.context._ctx)
        self.efx.alDeleteFilters(1, &self.id)
        alc.alcMakeContextCurrent(prev_ctx)

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, val):
        cdef str val_upper = val.upper()
        cdef enum_val = al.alGetEnumValue(b"AL_FILTER_" + val_upper.encode("utf8"))
        if enum_val == al.AL_NONE:
            raise RuntimeError(f"Could not get enum value for AL_FILTER_{val_upper}")
        self.efx.alFilteri(self.id, self.efx.AL_FILTER_TYPE, enum_val)
        check_al_error()
        self._type = val

    def get_int(self, prop):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALint val
        self.efx.alGetFilteri(self.id, e_val, &val)
        check_al_error()
        return val

    def get_float(self, prop):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALfloat val
        self.efx.alGetFilterf(self.id, e_val, &val)
        check_al_error()
        return val

    def get_v3f(self, prop):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALfloat[3] val
        self.efx.alGetfilterfv(self.id, e_val, val)
        check_al_error()
        return V3f(val[0], val[1], val[2])

    def set(self, prop, val):
        cdef al.ALenum e_val = get_al_enum(self._type + "_" + prop)
        cdef al.ALfloat[3] data
        if isinstance(val, int):
            self.efx.alFilteri(self.id, e_val, val)
        elif isinstance(val, float):
            self.efx.alFilterf(self.id, e_val, val)
        elif (isinstance(val, Sequence) and len(val) == 3) or isinstance(val, V3f):
            data = [val[0], val[1], val[2]]
            self.efx.alFilterfv(self.id, e_val, data)
        else:
            raise TypeError("Cannot convert to OpenAL type")
        check_al_error()

cdef object make_effect(type cls, EfxExtension efx, al.ALuint id, dict kwargs):
    cdef Effect effect = <Effect>cls.__new__(cls)
    effect.init_with_id(efx, id)
    for k, v in kwargs.items(): effect.set(k, v)
    return effect

cdef object make_filter(type cls, EfxExtension efx, al.ALuint id, dict kwargs):
    cdef Filter filter = <Filter>cls.__new__(cls)
    filter.init_with_id(efx, id)
    for k, v in kwargs.items(): filter.set(k, v)
    return filter



# cython: language_level=3

from .buffer cimport Buffer
from .context cimport Context
from . cimport al

cdef class Source:
    cdef object __weakref__
    cdef readonly Context context
    cdef Buffer _buf
    cdef dict _queued_bufs
    cdef readonly al.ALuint id

    @staticmethod
    cdef Source from_id(Context ctx, al.ALuint id)



# cython: language_level=3

from .device cimport Device
from .listener cimport Listener
from . cimport al, alc

cdef class Context:
    cdef object __weakref__
    cdef alc.ALCcontext* _ctx
    cdef readonly Device device
    cdef readonly Listener listener

    # Flags for extension emulation
    cdef bint emulate_direct_channels
    cdef bint emulate_direct_channels_remix
    cdef bint emulate_source_spatialize
    cdef bint emulate_disconnect
    cdef bint emulate_buffer_length_query

    # Extension enum values

    # From ALC_EXT_DISCONNECT
    cdef al.ALenum alc_connected
    
    # From AL_SOFT_buffer_length_query
    cdef al.ALenum al_byte_length_soft
    cdef al.ALenum al_sample_length_soft
    cdef al.ALenum al_sec_length_soft

    # Extension function pointers

    # From AL_SOFT_deferred_updates
    cdef void (*al_defer_updates_soft)()
    cdef void (*al_process_updates_soft)()

cdef class ContextAttrs:
    cdef object __weakref__
    cdef alc.ALCint[:] _attrs
    cdef Device _dev



# cython: language_level=3

from .exceptions cimport UnsupportedExtensionError, check_al_error
from . cimport al, alc

cdef class EventsExtension:
    def __cinit__(self):
        if al.alIsExtensionPresent(b"AL_SOFT_EVENTS") == al.AL_FALSE:
            check_al_error()
            raise UnsupportedExtensionError("ALC_SOFT_EVENTS")



""" Capture 5 seconds of audio at 48000 hz, using the OpenAL capture extension,
and writing it to a wave file with the builtin wave module.
"""

import cyal, wave

FREQ = 48000

ext = cyal.CaptureExtension()
dev = ext.open_device(sample_rate=FREQ)
total = 0
with wave.open("out.wav", "w") as w:
    w.setnchannels(1)
    w.setsampwidth(2)
    w.setframerate(FREQ)
    with dev.capturing():
        while total < FREQ * 5:
            bytes = dev.available_samples * 2
            print(bytes)
            buf = bytearray(bytes)
            dev.capture_samples(buf)
            w.writeframes(buf)
            total += bytes



# cython: language_level=3

from libc.limits cimport INT_MAX
from cpython cimport array
import array
from .context cimport Context
from .exceptions cimport check_al_error, UnsupportedExtensionError
from .util cimport V3f
from . cimport al, alc

cdef array.array ids_template = array.array('I')

cdef class Source:
    def __cinit__(self):
        pass

    def __init__(self):
        raise TypeError("This class cannot be instantiated directly.")

    @staticmethod
    cdef Source from_id(Context ctx, al.ALuint id):
        cdef Source src = Source.__new__(Source)
        src.context = ctx
        src._queued_bufs = {}
        src.id = id
        return src

    def __dealloc__(self):
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(self.context._ctx)
        al.alDeleteSources(1, &self.id)
        alc.alcMakeContextCurrent(prev_ctx)

    def play(self):
        al.alSourcePlay(self.id)
        check_al_error()

    def stop(self):
        al.alSourceStop(self.id)
        check_al_error()

    def rewind(self):
        al.alSourceRewind(self.id)
        check_al_error()

    def pause(self):
        al.alSourcePause(self.id)
        check_al_error()

    def queue_buffers(self, *bufs):
        cdef al.ALuint[:] ids = array.clone(ids_template, len(bufs), zero=False)
        cdef Py_ssize_t i
        for i, b in enumerate(bufs): ids[i] = b.id
        al.alSourceQueueBuffers(self.id, ids.size, &ids[0])
        check_al_error()
        for b in bufs: self._queued_bufs[b.id] = b
        self._buf = None

    def unqueue_buffers(self, *, max=INT_MAX):
        cdef al.ALint length
        al.alGetSourcei(self.id, al.AL_BUFFERS_PROCESSED, &length)
        length = length if length < max else max
        if length <= 0:
            return []
        cdef al.ALuint[:] ids = array.clone(ids_template, length, zero=False)
        al.alSourceUnqueueBuffers(self.id, ids.size, &ids[0])
        check_al_error()
        return [self._queued_bufs.pop(id) for id in ids]

    @property
    def buffer(self):
        return self._buf

    @buffer.setter
    def buffer(self, buf):
        al.alSourcei(self.id, al.AL_BUFFER, <al.ALint>buf.id)
        check_al_error()
        self._buf = buf


    @buffer.deleter
    def buffer(self):
        al.alSourcei(self.id, al.AL_BUFFER, al.AL_NONE)
        self._buf = None

    @property
    def relative(self):
        cdef al.ALint flag
        al.alGetSourcei(self.id, al.AL_SOURCE_RELATIVE, &flag)
        return flag == al.AL_TRUE

    @relative.setter
    def relative(self, val):
        al.alSourcei(self.id, al.AL_SOURCE_RELATIVE, al.AL_TRUE if val else al.AL_FALSE)

    @property
    def cone_inner_angle(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_CONE_INNER_ANGLE, &val)
        return val

    @cone_inner_angle.setter
    def cone_inner_angle(self, val):
        al.alSourcef(self.id, al.AL_CONE_INNER_ANGLE, val)

    @property
    def cone_outer_angle(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_CONE_OUTER_ANGLE, &val)
        return val

    @cone_outer_angle.setter
    def cone_outer_angle(self, val):
        al.alSourcef(self.id, al.AL_CONE_OUTER_ANGLE, val)

    @property
    def pitch(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_PITCH, &val)
        return val

    @pitch.setter
    def pitch(self, val):
        al.alSourcef(self.id, al.AL_PITCH, val)

    @property
    def position(self):
        cdef float x, y, z
        al.alGetSource3f(self.id, al.AL_POSITION, &x, &y, &z)
        return V3f(x, y, z)

    @position.setter
    def position(self, val):
        al.alSource3f(self.id, al.AL_POSITION, val[0], val[1], val[2])

    @property
    def direction(self):
        cdef float x, y, z
        al.alGetSource3f(self.id, al.AL_DIRECTION, &x, &y, &z)
        return V3f(x, y, z)

    @direction.setter
    def direction(self, val):
        al.alSource3f(self.id, al.AL_DIRECTION, val[0], val[1], val[2])

    @property
    def velocity(self):
        cdef float x, y, z
        al.alGetSource3f(self.id, al.AL_VELOCITY, &x, &y, &z)
        return V3f(x, y, z)

    @velocity.setter
    def velocity(self, val):
        al.alSource3f(self.id, al.AL_VELOCITY, val[0], val[1], val[2])

    @property
    def looping(self):
        cdef al.ALint flag
        al.alGetSourcei(self.id, al.AL_LOOPING, &flag)
        return flag == al.AL_TRUE

    @looping.setter
    def looping(self, val):
        al.alSourcei(self.id, al.AL_LOOPING, al.AL_TRUE if val else al.AL_FALSE)

    @property
    def gain(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_GAIN, &val)
        return val

    @gain.setter
    def gain(self, val):
        al.alSourcef(self.id, al.AL_GAIN, val)

    @property
    def min_gain(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_MIN_GAIN, &val)
        return val

    @min_gain.setter
    def min_gain(self, val):
        al.alSourcef(self.id, al.AL_MIN_GAIN, val)

    @property
    def max_gain(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_MAX_GAIN, &val)
        return val

    @max_gain.setter
    def max_gain(self, val):
        al.alSourcef(self.id, al.AL_MAX_GAIN, val)

    @property
    def state(self):
        cdef al.ALenum val
        al.alGetSourcei(self.id, al.AL_SOURCE_STATE, &val)
        return SourceState(val)

    @property
    def buffers_queued(self):
        cdef al.ALint val
        al.alGetSourcei(self.id, al.AL_BUFFERS_QUEUED, &val)
        return val

    @property
    def buffers_processed(self):
        cdef al.ALint val
        al.alGetSourcei(self.id, al.AL_BUFFERS_PROCESSED, &val)
        return val

    @property
    def reference_distance(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_REFERENCE_DISTANCE, &val)
        return val

    @reference_distance.setter
    def reference_distance(self, val):
        al.alSourcef(self.id, al.AL_REFERENCE_DISTANCE, val)

    @property
    def rolloff_factor(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_ROLLOFF_FACTOR, &val)
        return val

    @rolloff_factor.setter
    def rolloff_factor(self, val):
        al.alSourcef(self.id, al.AL_ROLLOFF_FACTOR, val)

    @property
    def cone_outer_gain(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_CONE_OUTER_GAIN, &val)
        return val

    @cone_outer_gain.setter
    def cone_outer_gain(self, val):
        al.alSourcef(self.id, al.AL_CONE_OUTER_GAIN, val)

    @property
    def max_distance(self):
        cdef float val
        al.alGetSourcef(self.id, al.AL_MAX_DISTANCE, &val)
        return val

    @max_distance.setter
    def max_distance(self, val):
        al.alSourcef(self.id, al.AL_MAX_DISTANCE, val)

    @property
    def type(self):
        cdef al.ALenum val
        al.alGetSourcei(self.id, al.AL_SOURCE_TYPE, &val)
        return SourceType(val)

    @property
    def direct_channels(self):
        cdef al.ALenum direct_channels_soft = al.alGetEnumValue(b"AL_DIRECT_CHANNELS_SOFT")
        cdef al.ALenum remix_unmatched_soft = al.alGetEnumValue(b"AL_REMIX_UNMATCHED_SOFT")
        cdef al.ALint val

        if direct_channels_soft != al.AL_NONE:
            al.alGetSourcei(self.id, direct_channels_soft, &val)
            check_al_error()

            if val == al.AL_FALSE: return False
            elif val == al.AL_TRUE: return True
            else:
                if remix_unmatched_soft != al.AL_NONE:
                    if val == remix_unmatched_soft: return "remix_unmatched"
                    else: raise ValueError(f"Invalid value for Source.direct_channels, got {val}")
                elif self.context.emulate_direct_channels_remix: return True
                else: raise ValueError(f"Invalid value for Source.direct_channels, got {val}")

        else:
            if self.context.emulate_direct_channels: return False
            else: raise UnsupportedExtensionError("AL_SOFT_DIRECT_CHANNELS")

    @direct_channels.setter
    def direct_channels(self, val):
        cdef al.ALenum direct_channels_soft = al.alGetEnumValue(b"AL_DIRECT_CHANNELS_SOFT")
        cdef al.ALenum remix_unmatched_soft = al.alGetEnumValue(b"AL_REMIX_UNMATCHED_SOFT")
        cdef al.ALint enum_val
        if direct_channels_soft != al.AL_NONE:
            if val == True: enum_val = al.AL_TRUE
            elif val == False: enum_val = al.AL_FALSE
            elif val == "remix_unmatched":
                if remix_unmatched_soft != al.AL_NONE: enum_val = remix_unmatched_soft
                elif self.context.emulate_direct_channels_remix: enum_value = al.AL_TRUE
                else: raise UnsupportedExtensionError("AL_SOFT_DIRECT_CHANNELS_REMIX")
            else: raise ValueError(f"Invalid value for Source.direct_channels, got {val}")
            al.alSourcei(self.id, direct_channels_soft, enum_val)
            check_al_error()
        elif not self.context.emulate_direct_channels:
            raise UnsupportedExtensionError("AL_SOFT_DIRECT_CHANNELS")

    @property
    def spatialize(self):
        cdef al.ALenum source_spatialize_soft = al.alGetEnumValue(b"AL_SOURCE_SPATIALIZE_SOFT")
        cdef al.ALint val
        if source_spatialize_soft != 0:
            al.alGetSourcei(self.id, source_spatialize_soft, &val)
            check_al_error()
            if val == al.AL_FALSE: return False
            elif val == al.AL_TRUE: return True
            elif val == al.alGetEnumValue(b"AL_AUTO_SOFT"): return "auto"
            else: raise ValueError(f"Invalid value for Source.spatialize, got {val}")
        else:
            if self.context.emulate_source_spatialize: return "auto" 
            else: raise UnsupportedExtensionError("AL_SOFT_SOURCE_SPATIALIZE")

    @spatialize.setter
    def spatialize(self, val):
        cdef al.ALenum source_spatialize_soft = al.alGetEnumValue(b"AL_SOURCE_SPATIALIZE_SOFT")
        cdef al.ALint enum_val
        if source_spatialize_soft != 0:
            if val == False: enum_val = al.AL_FALSE
            elif val == True: enum_val = al.AL_TRUE
            elif val == "auto": enum_val = al.alGetEnumValue(b"AL_AUTO_SOFT")
            else: raise ValueError(f"Invalid value for Source.spatialize, got {val}")
            al.alSourcei(self.id, source_spatialize_soft, enum_val)
            check_al_error()
        elif not self.context.emulate_source_spatialize:
            raise UnsupportedExtensionError("AL_SOFT_SOURCE_SPATIALIZE")

cpdef enum SourceState:
    INITIAL = al.AL_INITIAL
    PLAYING = al.AL_PLAYING
    PAUSED = al.AL_PAUSED
    STOPPED = al.AL_STOPPED

cpdef enum SourceType:
        STATIC = al.AL_STATIC
        STREAMING = al.AL_STREAMING
        UNDETERMINED = al.AL_UNDETERMINED



# cython: language_level=3

from . cimport al, alc
from .context cimport Context

cdef class EfxExtension:
    cdef object __weakref__
    cdef readonly Context context

    # Function pointers
    cdef void (*alGenEffects)(al.ALsizei n, al.ALuint *effects)
    cdef void (*alDeleteEffects)(al.ALsizei n, const al.ALuint *effects)
    cdef void (*alEffecti)(al.ALuint effect, al.ALenum param, al.ALint iValue)
    cdef void (*alEffectiv)(al.ALuint effect, al.ALenum param, const al.ALint *piValues)
    cdef void (*alEffectf)(al.ALuint effect, al.ALenum param, al.ALfloat flValue)
    cdef void (*alEffectfv)(al.ALuint effect, al.ALenum param, const al.ALfloat *pflValues)
    cdef void (*alGetEffecti)(al.ALuint effect, al.ALenum param, al.ALint *piValue)
    cdef void (*alGetEffectiv)(al.ALuint effect, al.ALenum param, al.ALint *piValues)
    cdef void (*alGetEffectf)(al.ALuint effect, al.ALenum param, al.ALfloat *pflValue)
    cdef void (*alGetEffectfv)(al.ALuint effect, al.ALenum param, al.ALfloat *pflValues)

    cdef void (*alGenFilters)(al.ALsizei n, al.ALuint *filters)
    cdef void (*alDeleteFilters)(al.ALsizei n, const al.ALuint *filters)
    cdef void (*alFilteri)(al.ALuint filter, al.ALenum param, al.ALint iValue)
    cdef void (*alFilteriv)(al.ALuint filter, al.ALenum param, const al.ALint *piValues)
    cdef void (*alFilterf)(al.ALuint filter, al.ALenum param, al.ALfloat flValue)
    cdef void (*alFilterfv)(al.ALuint filter, al.ALenum param, const al.ALfloat *pflValues)
    cdef void (*alGetFilteri)(al.ALuint filter, al.ALenum param, al.ALint *piValue)
    cdef void (*alGetFilteriv)(al.ALuint filter, al.ALenum param, al.ALint *piValues)
    cdef void (*alGetFilterf)(al.ALuint filter, al.ALenum param, al.ALfloat *pflValue)
    cdef void (*alGetFilterfv)(al.ALuint filter, al.ALenum param, al.ALfloat *pflValues)

    cdef void (*alGenAuxiliaryEffectSlots)(al.ALsizei n, al.ALuint *effectslots)
    cdef void (*alDeleteAuxiliaryEffectSlots)(al.ALsizei n, const al.ALuint *effectslots)
    cdef void (*alAuxiliaryEffectSloti)(al.ALuint effectslot, al.ALenum param, al.ALint iValue)
    cdef void (*alAuxiliaryEffectSlotiv)(al.ALuint effectslot, al.ALenum param, const al.ALint *piValues)
    cdef void (*alAuxiliaryEffectSlotf)(al.ALuint effectslot, al.ALenum param, al.ALfloat flValue)
    cdef void (*alAuxiliaryEffectSlotfv)(al.ALuint effectslot, al.ALenum param, const al.ALfloat *pflValues)
    cdef void (*alGetAuxiliaryEffectSloti)(al.ALuint effectslot, al.ALenum param, al.ALint *piValue)
    cdef void (*alGetAuxiliaryEffectSlotiv)(al.ALuint effectslot, al.ALenum param, al.ALint *piValues)
    cdef void (*alGetAuxiliaryEffectSlotf)(al.ALuint effectslot, al.ALenum param, al.ALfloat *pflValue)
    cdef void (*alGetAuxiliaryEffectSlotfv)(al.ALuint effectslot, al.ALenum param, al.ALfloat *pflValues)

    # Enum values
    cdef al.ALenum AL_METERS_PER_UNIT
    cdef alc.ALCenum alc_max_auxiliary_sends
    cdef al.ALenum AL_EFFECT_TYPE
    cdef al.ALenum AL_FILTER_TYPE

cdef class AuxiliaryEffectSlot:
    cdef object __weakref__
    cdef readonly EfxExtension efx
    cdef readonly al.ALuint id

    @staticmethod
    cdef AuxiliaryEffectSlot from_id(EfxExtension efx, al.ALuint id)

cdef class Effect:
    cdef object __weakref__
    cdef readonly EfxExtension efx
    cdef readonly al.ALuint id
    cdef str _type

    cdef void init_with_id(self, EfxExtension efx, al.ALuint id)

cdef class Filter:
    cdef object __weakref__
    cdef readonly EfxExtension efx
    cdef readonly al.ALuint id
    cdef str _type

    cdef void init_with_id(self, EfxExtension efx, al.ALuint id)



from .about import __doc__, __version__, __author__, __author_email__
from .buffer import Buffer, BufferFormat
from .capture import CaptureExtension, CaptureDevice
from .context import Context
from .device import Device
from .source import Source, SourceState, SourceType
from .util import get_device_specifiers, get_default_device_specifier, get_all_device_specifiers, get_default_all_device_specifier, get_supported_extensions, get_version, V3f



# cython: language_level=3

from contextlib import contextmanager

from .context cimport ContextAttrs
from .exceptions cimport DeviceNotFoundError, UnsupportedExtensionError, check_alc_error
from . cimport al, alc

cdef class Device:
    def __cinit__(self, name=None, *, pause_noop=False, emulate_enumerate_all=True):
        cdef bytes name_bytes = name.encode("utf8") if name is not None else b''
        self._device = alc.alcOpenDevice(<const alc.ALCchar*>name_bytes if name is not None else NULL)
        if self._device is NULL:
            raise DeviceNotFoundError(device_name=name)
        self.get_al_proc_address = <al.ALvoid*(*)(const al.ALchar*)>self.get_alc_proc_address("alGetProcAddress")
        if alc.alcIsExtensionPresent(self._device, "ALC_SOFT_PAUSE_DEVICE") == al.AL_TRUE:
            self.pause_soft = <void (*)(alc.ALCdevice*)>self.get_alc_proc_address("alcDevicePauseSOFT")
            self.resume_soft = <void (*)(alc.ALCdevice*)>self.get_alc_proc_address("alcDeviceResumeSOFT")
        elif pause_noop:
            self.pause_soft = do_nothing_with_device
            self.resume_soft = do_nothing_with_device
        else:
            self.pause_soft = no_pause_device_ext
            self.resume_soft = no_pause_device_ext
        self.alc_reset_device_soft = <alc.ALCboolean (*)(alc.ALCdevice*, alc.ALCint*)>self.get_alc_proc_address(b"alcResetDeviceSOFT")
        self.alc_reopen_device_soft = <alc.ALCboolean (*)(alc.ALCdevice*, const alc.ALCchar*, alc.ALCint*)>self.get_alc_proc_address(b"alcReopenDeviceSOFT")

        # Flags for extension emulation
        self.emulate_enumerate_all = emulate_enumerate_all
    
    def __dealloc__(self):
        if self._device is not NULL:
            alc.alcCloseDevice(self._device)

    def pause(self):
        self.pause_soft(self._device)

    def resume(self):
        self.resume_soft(self._device)

    @contextmanager
    def paused(self):
        self.pause_soft(self._device)
        try:
            yield self
        finally:
            self.resume_soft(self._device)

    @contextmanager
    def playing(self):
        self.resume_soft(self._device)
        try:
            yield self
        finally:
            self.pause_soft(self._device)

    @property
    def name(self):
        return alc.alcGetString(self._device, alc.ALC_DEVICE_SPECIFIER).decode("utf8")

    @property
    def output_name(self):
        cdef alc.ALCenum enum_val = alc.alcGetEnumValue(NULL, b"ALC_ALL_DEVICES_SPECIFIER")
        cdef const alc.ALCchar *spec
        if enum_val != al.AL_NONE:
            spec = alc.alcGetString(self._device, enum_val)
            return spec.decode("utf8")
        elif self.emulate_enumerate_all:
            return self.name
        else:
            raise UnsupportedExtensionError("ALC_ENUMERATE_ALL_EXT")

    @property
    def version(self):
        cdef alc.ALCint major, minor
        alc.alcGetIntegerv(self._device, alc.ALC_MAJOR_VERSION, 1, &major)
        alc.alcGetIntegerv(self._device, alc.ALC_MINOR_VERSION, 1, &minor)
        return (major, minor)

    def reset(self, **kwargs):
        if self.alc_reopen_device_soft is NULL:
            raise UnsupportedExtensionError("ALC_SOFT_HRTF or ALC_SOFT_OUTPUT_LIMITER")
        cdef ContextAttrs attrs = ContextAttrs.from_kwargs(self, **kwargs)
        cdef alc.ALCboolean res = self.alc_reset_device_soft(self._device, &attrs._attrs[0])
        if res != al.AL_TRUE:
            check_alc_error(self._device)

    def reopen(self, name=None, **kwargs):
        if self.alc_reopen_device_soft is NULL:
            raise UnsupportedExtensionError("ALC_SOFT_REOPEN_DEVICE")
        cdef bytes name_bytes = name.encode("utf8") if name is not None else b''
        cdef ContextAttrs attrs = ContextAttrs.from_kwargs(self, **kwargs)
        cdef alc.ALCboolean res = self.alc_reopen_device_soft(self._device, <const alc.ALCchar*>name_bytes, &attrs._attrs[0])
        if res != al.AL_TRUE:
            check_alc_error(self._device)

    cpdef list get_supported_extensions(self):
        return alc.alcGetString(self._device, alc.ALC_EXTENSIONS).decode("utf8").split(' ')

    def is_extension_present(self, ext):
        return alc.alcIsExtensionPresent(self._device, <const alc.ALCchar *>ext) == al.AL_TRUE

    cdef alc.ALCenum get_enum_value(self, const alc.ALCchar* name):
        cdef alc.ALCenum val = alc.alcGetEnumValue(self._device, name)
        if val == al.AL_NONE:
            raise RuntimeError(f"Could not get enum value for {name}")
        return val

cdef void no_pause_device_ext(alc.ALCdevice* dev):
    raise UnsupportedExtensionError("ALC_SOFT_PAUSE_DEVICE")

cdef void do_nothing_with_device(alc.ALCdevice* dev): pass



# cython: language_level=3

from .context cimport Context
from .exceptions cimport check_al_error, UnsupportedExtensionError
from . cimport al, alc

cdef class Buffer:
    def __cinit__(self):
        pass

    def __init__(self):
        raise TypeError("This class cannot be instantiated directly.")

    @staticmethod
    cdef Buffer from_id(Context ctx, al.ALuint id):
        cdef Buffer buf = Buffer.__new__(Buffer)
        buf.context = ctx
        buf.id = id
        return buf

    def __dealloc__(self):
        cdef alc.ALCcontext* prev_ctx = alc.alcGetCurrentContext()
        alc.alcMakeContextCurrent(self.context._ctx)
        al.alDeleteBuffers(1, &self.id)
        alc.alcMakeContextCurrent(prev_ctx)

    def set_data(self, data, *, sample_rate, format):
        cdef const al.ALubyte[:] view = data
        al.alBufferData(self.id, format, &view[0], view.size, sample_rate)
        check_al_error()

    def __len__(self):
        cdef al.ALsizei length
        al.alGetBufferi(self.id, al.AL_SIZE, &length)
        return length

    @property
    def bits(self):
        cdef al.ALint val
        al.alGetBufferi(self.id, al.AL_BITS, &val)
        return val

    @property
    def channels(self):
        cdef al.ALint val
        al.alGetBufferi(self.id, al.AL_CHANNELS, &val)
        return val

    @property
    def sample_rate(self):
        cdef al.ALsizei val
        al.alGetBufferi(self.id, al.AL_FREQUENCY, &val)
        return val

    @property
    def byte_length(self):
        cdef al.ALint val
        if self.context.al_byte_length_soft != al.AL_NONE:
            al.alGetBufferi(self.id, self.context.al_byte_length_soft, &val)
            return val
        elif self.context.emulate_buffer_length_query:
            return len(self)
        else:
            raise UnsupportedExtensionError("AL_SOFT_BUFFER_LENGTH_QUERY")

    @property
    def sample_length(self):
        cdef al.ALint val
        if self.context.al_sample_length_soft != al.AL_NONE:
            al.alGetBufferi(self.id, self.context.al_sample_length_soft, &val)
            return val
        elif self.context.emulate_buffer_length_query:
            return len(self) // (self.bits // 8) // self.channels
        else:
            raise UnsupportedExtensionError("AL_SOFT_BUFFER_LENGTH_QUERY")

    @property
    def sec_length(self):
        cdef al.ALfloat val
        if self.context.al_sec_length_soft != al.AL_NONE:
            al.alGetBufferf(self.id, self.context.al_sec_length_soft, &val)
            return val
        elif self.context.emulate_buffer_length_query:
            try: return len(self) // (self.bits // 8) // self.channels / self.sample_rate
            except ZeroDivisionError: return 0
        else:
            raise UnsupportedExtensionError("AL_SOFT_BUFFER_LENGTH_QUERY")
